import math
import torch

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Any
from typing import List
from typing import Tuple
from typing import Optional

from .protocol import Decoder1DBase
from ....types import tensor_dict_type
from ....protocol import TrainerState
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....constants import PREDICTIONS_KEY
from ....modules.blocks import Conv2d
from ....modules.blocks import Activation
from ....modules.blocks import UpsampleConv2d


def setup_filter(
    f: Any,
    device: torch.device = torch.device("cpu"),
    normalize: bool = True,
    flip_filter: bool = False,
    gain: float = 1.0,
    separable: Optional[bool] = None,
) -> Tensor:
    if f is None:
        f = 1
    f = torch.as_tensor(f, dtype=torch.float32)
    assert f.ndim in [0, 1, 2]
    assert f.numel() > 0
    if f.ndim == 0:
        f = f[np.newaxis]
    if separable is None:
        separable = f.ndim == 1 and f.numel() >= 8
    if f.ndim == 1 and not separable:
        f = f.ger(f)
    assert f.ndim == (1 if separable else 2)
    if normalize:
        f /= f.sum()
    if flip_filter:
        f = f.flip(list(range(f.ndim)))
    f = f * (gain ** (f.ndim / 2))
    f = f.to(device=device)
    return f


def resample(
    net: Tensor,
    f: Tensor,
    upscale: float,
    padding: List[int],
    gain: float,
) -> Tensor:
    c, h, w = map(int, net.shape[1:])
    upscale = int(upscale)
    net = net.reshape([-1, c, h, 1, w, 1])
    net = F.pad(net, [0, upscale - 1, 0, 0, 0, upscale - 1])
    net = net.reshape([-1, c, h * upscale, w * upscale])
    net = F.pad(net, padding)
    f = f * (gain ** (f.ndim / 2))
    f = f[None, None, ...]
    f = f.repeat([c, 1] + [1] * (f.ndim - 2))
    net = F.conv2d(net, f, groups=c)
    return net


class FullyConnected(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        bias: bool = True,
        lr_multiplier: float = 1.0,
        bias_init: float = 0.0,
        activation: Optional[str] = None,
        activation_gain: float = math.sqrt(2.0),
    ):
        super().__init__()
        self.weight = nn.Parameter(torch.randn([out_dim, in_dim]) / lr_multiplier)
        self.bias = nn.Parameter(torch.full([out_dim], bias_init)) if bias else None
        self.weight_gain = lr_multiplier / np.sqrt(in_dim)
        self.bias_gain = lr_multiplier
        if activation is None:
            self.activation = None
        else:
            self.activation = Activation.make(activation)
        self.activation_gain = activation_gain

    def forward(self, net: Tensor) -> Tensor:
        w = self.weight * self.weight_gain
        b = self.bias
        if b is None:
            net = net.matmul(w.t())
        if b is not None:
            if self.bias_gain != 1:
                b = b * self.bias_gain
            net = torch.addmm(b.unsqueeze(0), net, w.t())
        if self.activation is not None:
            net = self.activation(net)
            net = net * self.activation_gain
        return net


class ToRGB(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        w_dim: int,
        kernel_size: int = 1,
        conv_clamp: Optional[float] = None,
    ):
        super().__init__()
        self.conv_clamp = conv_clamp
        self.affine = FullyConnected(w_dim, in_channels, bias_init=1.0)
        self.conv = Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            bias=False,
        )
        self.bias = nn.Parameter(torch.zeros([1, out_channels, 1, 1]))
        self.weight_gain = 1.0 / np.sqrt(in_channels * (kernel_size**2))

    def forward(self, net: Tensor, w: Tensor) -> Tensor:
        styles = self.affine(w) * self.weight_gain
        net = self.conv(net, styles) + self.bias
        if self.conv_clamp is not None:
            net = torch.clamp(net, -self.conv_clamp, self.conv_clamp)
        return net


class StyleGAN2Conv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        latent_dim: int,
        resolution: int,
        kernel_size: int = 3,
        upscale_factor: float = 1.0,
        use_noise: bool = True,
        activation: str = "leaky_relu_0.2",
        activation_gain: float = math.sqrt(2.0),
        resample_filter: Optional[List[int]] = None,
        conv_clamp: Optional[float] = None,
    ):
        super().__init__()
        if resample_filter is None:
            resample_filter = [1, 3, 3, 1]
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.resolution = resolution
        self.upscale_factor = upscale_factor
        self.use_noise = use_noise
        self.activation = Activation.make(activation)
        self.activation_gain = activation_gain
        self.conv_clamp = conv_clamp
        self.register_buffer("resample_filter", setup_filter(resample_filter))
        self.affine = FullyConnected(latent_dim, in_channels, bias_init=1.0)
        self.conv = UpsampleConv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            mode="transpose",
            bias=False,
            demodulate=True,
            factor=upscale_factor,
        )
        if use_noise:
            self.register_buffer("noise_const", torch.randn([resolution, resolution]))
            self.noise_strength = nn.Parameter(torch.tensor(0.0))
        self.bias = nn.Parameter(torch.zeros([1, out_channels, 1, 1]))

    def forward(
        self,
        net: Tensor,
        w: Tensor,
        noise_mode: str = "random",
        gain: float = 1.0,
    ) -> Tensor:
        styles = self.affine(w)
        net = self.conv(net, styles)
        if self.upscale_factor > 1.0:
            net = resample(
                net,
                self.resample_filter,
                1.0,
                [1] * 4,
                self.upscale_factor**2,
            )
        if self.use_noise:
            if noise_mode == "const":
                noise = self.noise_const
            elif noise_mode == "random":
                noise_shape = net.shape[0], 1, self.resolution, self.resolution
                noise = torch.randn(noise_shape, device=net.device)
            else:
                raise NotImplementedError(f"noise_mode '{noise_mode}' not implemented")
            noise = noise * self.noise_strength
            net.add_(noise)
        net = net + self.bias
        activation_gain = self.activation_gain * gain
        if activation_gain != 1.0:
            net = net * activation_gain
        net = self.activation(net)
        if self.conv_clamp is not None:
            conv_clamp = self.conv_clamp * gain
            net = torch.clamp(net, -conv_clamp, conv_clamp)
        return net


class StyleGAN2Block(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        latent_dim: int,
        resolution: int,
        img_channels: int,
        is_last: bool,
        architecture: str = "skip",
        resample_filter: Optional[List[int]] = None,
        conv_clamp: Optional[float] = None,
        **conv_kwargs: Any,
    ):
        super().__init__()
        if resample_filter is None:
            resample_filter = [1, 3, 3, 1]
        self.in_channels = in_channels
        self.w_dim = latent_dim
        self.resolution = resolution
        self.img_channels = img_channels
        self.is_last = is_last
        self.architecture = architecture
        self.register_buffer("resample_filter", setup_filter(resample_filter))
        self.num_conv = 0
        self.num_to_rgb = 0

        if in_channels == 0:
            self.conv0 = None
        else:
            self.conv0 = StyleGAN2Conv(
                in_channels,
                out_channels,
                latent_dim=latent_dim,
                resolution=resolution,
                upscale_factor=2.0,
                resample_filter=resample_filter,
                conv_clamp=conv_clamp,
                **conv_kwargs,
            )
            self.num_conv += 1

        self.conv1 = StyleGAN2Conv(
            out_channels,
            out_channels,
            latent_dim=latent_dim,
            resolution=resolution,
            conv_clamp=conv_clamp,
            **conv_kwargs,
        )
        self.num_conv += 1

        if not (is_last or architecture == "skip"):
            self.to_rgb = None
        else:
            self.to_rgb = ToRGB(
                out_channels,
                img_channels,
                w_dim=latent_dim,
                conv_clamp=conv_clamp,
            )
            self.num_to_rgb += 1

    def forward(
        self,
        net: Optional[Tensor],
        rgb: Optional[Tensor],
        ws: Tensor,
        *,
        determinate: bool = False,
        **conv_kwargs: Any,
    ) -> Tuple[Tensor, Tensor]:
        w_iter = iter(ws.unbind(dim=1))
        if determinate:
            conv_kwargs.setdefault("noise_mode", "const")
        if self.conv0 is not None:
            net = self.conv0(net, next(w_iter), **conv_kwargs)
        net = self.conv1(net, next(w_iter), **conv_kwargs)
        if rgb is not None:
            rgb = resample(rgb, self.resample_filter, 2.0, [2, 1, 2, 1], 4.0)
        if self.to_rgb is not None:
            y = self.to_rgb(net, next(w_iter))
            y = y.to(dtype=torch.float32, memory_format=torch.contiguous_format)
            rgb = rgb.add_(y) if rgb is not None else y
        return net, rgb


class StyleGAN2Decoder(nn.Module):
    def __init__(
        self,
        img_size: int,
        latent_dim: int = 512,
        out_channels: int = 3,
        *,
        channel_base: int = 32768,
        max_channels: int = 512,
        num_classes: Optional[int] = None,
        conv_clamp: Optional[float] = 256.0,
        **block_kwargs: Any,
    ):
        assert img_size >= 4 and img_size & (img_size - 1) == 0
        super().__init__()
        self.w_dim = latent_dim
        self.img_size = img_size
        self.img_size_log2 = int(round(math.log2(img_size)))
        self.img_channels = out_channels
        self.block_resolutions = [2**i for i in range(2, self.img_size_log2 + 1)]
        channels_dict = {
            resolution: min(channel_base // resolution, max_channels)
            for resolution in self.block_resolutions
        }
        self.out_channels = out_channels
        self.num_upsample = len(self.block_resolutions)
        self.latent_resolution = self.block_resolutions[0]
        self.latent_channels = channels_dict[self.latent_resolution]
        self.num_classes = num_classes
        shape = (1 if num_classes is None else num_classes), self.latent_channels, 4, 4
        self.const = nn.Parameter(torch.randn(*shape))
        blocks = {}
        self.num_ws = 0
        for resolution in self.block_resolutions:
            in_nc = channels_dict[resolution // 2] if resolution > 4 else 0
            out_nc = channels_dict[resolution]
            is_last = resolution == self.img_size
            block = StyleGAN2Block(
                in_nc,
                out_nc,
                latent_dim=latent_dim,
                resolution=resolution,
                img_channels=out_channels,
                is_last=is_last,
                conv_clamp=conv_clamp,
                **block_kwargs,
            )
            self.num_ws += block.num_conv
            if is_last:
                self.num_ws += block.num_to_rgb
            blocks[str(resolution)] = block
        self.blocks = nn.ModuleDict(blocks)

    def forward(self, ws: Tensor, *, labels: Optional[Tensor], **kwargs: Any) -> Tensor:
        w_idx = 0
        block_ws = []
        for resolution in self.block_resolutions:
            block = self.blocks[str(resolution)]
            block_ws.append(ws.narrow(1, w_idx, block.num_conv + block.num_to_rgb))
            w_idx += block.num_conv
        rgb = None
        batch_size = ws.shape[0]
        if self.num_classes is None:
            net = self.const.repeat([batch_size, 1, 1, 1])
        else:
            if labels is None:
                labels = torch.randint(self.num_classes, [batch_size], device=ws.device)
            net = self.const[labels.view(-1)]
        for resolution, resolution_ws in zip(self.block_resolutions, block_ws):
            block = self.blocks[str(resolution)]
            net, rgb = block(net, rgb, resolution_ws, **kwargs)
        return rgb


@Decoder1DBase.register("style2")
class StyleDecoder2(Decoder1DBase):
    def __init__(
        self,
        img_size: int,
        latent_dim: int = 512,
        out_channels: int = 3,
        *,
        channel_base: int = 32768,
        max_channels: int = 512,
        num_classes: Optional[int] = None,
        conv_clamp: Optional[float] = 256.0,
        **block_kwargs: Any,
    ):
        assert img_size >= 4 and img_size & (img_size - 1) == 0
        num_upsample = int(round(math.log2(img_size))) - 1
        super().__init__(
            latent_dim,
            out_channels,
            img_size=img_size,
            num_upsample=num_upsample,
            num_classes=num_classes,
            latent_resolution=4,
        )
        block_kwargs.pop("num_upsample", None)
        block_kwargs.pop("latent_resolution", None)
        self.decoder = StyleGAN2Decoder(
            img_size,
            latent_dim,
            out_channels,
            channel_base=channel_base,
            max_channels=max_channels,
            num_classes=num_classes,
            conv_clamp=conv_clamp,
            **block_kwargs,
        )

    def _z2ws(self, z: Tensor) -> Tensor:
        return z.unsqueeze(1).repeat(1, self.decoder.num_ws, 1)

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        ws = self._z2ws(batch[INPUT_KEY])
        net = self.decoder(ws, labels=batch.get(LABEL_KEY), **kwargs)
        return {PREDICTIONS_KEY: net}


__all__ = [
    "FullyConnected",
    "StyleGAN2Decoder",
    "StyleDecoder2",
]



STYLE_KEY = "style"
STYLE_LATENTS_KEY = "style_latents"
CONTENT_LATENT_KEY = "content_latent"
STYLIZED_STYLE_LATENTS_KEY = "stylized_style_latents"
STYLIZED_CONTENT_LATENT_KEY = "stylized_content_latent"

INPUT_B_KEY = "input_b"
LABEL_B_KEY = "labels_b"



from torch import nn
from torch import Tensor
from typing import Any
from typing import Dict
from typing import Optional
from cfcv.misc.toolkit import clip_normalize

from .constants import STYLE_KEY
from .constants import STYLE_LATENTS_KEY
from .constants import CONTENT_LATENT_KEY
from .constants import STYLIZED_STYLE_LATENTS_KEY
from .constants import STYLIZED_CONTENT_LATENT_KEY
from ..decoder import VanillaDecoder
from ..encoder import BackboneEncoder
from ....types import losses_type
from ....types import tensor_dict_type
from ....protocol import TrainerState
from ....protocol import LossProtocol
from ....protocol import ModelProtocol
from ....constants import LOSS_KEY
from ....constants import INPUT_KEY
from ....constants import LATENT_KEY
from ....constants import PREDICTIONS_KEY
from ....misc.toolkit import adain_with_tensor
from ....misc.toolkit import mean_std
from ....misc.toolkit import interpolate


@ModelProtocol.register("adain")
class AdaINStylizer(ModelProtocol):
    def __init__(
        self,
        in_channels: int = 3,
        *,
        backbone: str = "vgg_style",
        backbone_config: Optional[Dict[str, Any]] = None,
        decoder_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.backbone = BackboneEncoder(
            backbone,
            in_channels,
            finetune=False,
            pretrained=True,
            backbone_config=backbone_config,
        )
        if decoder_config is None:
            decoder_config = {}
        decoder_config.setdefault("norm_type", None)
        decoder_config.setdefault("activation", "relu")
        self.decoder = VanillaDecoder(
            self.backbone.latent_channels,
            in_channels,
            num_repeats="repeated",
            num_upsample=self.backbone.num_downsample,
            **(decoder_config or {}),
        )

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        style = batch[STYLE_KEY]
        content = batch[INPUT_KEY]
        determinate = kwargs.pop("determinate", False)
        need_stylized_features = kwargs.pop("need_stylized_features", True)
        style_feats = self.backbone(batch_idx, {INPUT_KEY: style}, state, **kwargs)
        style_latent = style_feats.pop(LATENT_KEY)
        content_feats = self.backbone(batch_idx, {INPUT_KEY: content}, state, **kwargs)
        content_latent = content_feats[LATENT_KEY]
        args = content_latent, style_latent
        encoded = adain_with_tensor(*args, determinate=determinate)
        style_weight = kwargs.get("style_weight", 1.0)
        encoded = style_weight * encoded + (1.0 - style_weight) * content_latent
        rs = self.decoder(
            batch_idx,
            {INPUT_KEY: encoded},
            state,
            determinate=determinate,
            **kwargs,
        )
        decoded = interpolate(rs[PREDICTIONS_KEY], anchor=content)
        if not need_stylized_features:
            decoded_feats = decoded_content_latent = None
        else:
            decoded_inp = {INPUT_KEY: decoded}
            decoded_feats = self.backbone(batch_idx, decoded_inp, state, **kwargs)
            decoded_content_latent = decoded_feats.pop(LATENT_KEY)
        return {
            PREDICTIONS_KEY: decoded,
            STYLE_LATENTS_KEY: style_feats,
            CONTENT_LATENT_KEY: encoded,
            STYLIZED_STYLE_LATENTS_KEY: decoded_feats,
            STYLIZED_CONTENT_LATENT_KEY: decoded_content_latent,
        }

    def onnx_forward(self, batch: tensor_dict_type) -> Any:
        return self.stylize(batch[INPUT_KEY], batch[STYLE_KEY], determinate=True)

    def stylize(self, net: Tensor, style: Tensor, **kwargs: Any) -> Tensor:
        inp = {INPUT_KEY: net, STYLE_KEY: style}
        kwargs["need_stylized_features"] = False
        decoded = self.forward(0, inp, **kwargs)[PREDICTIONS_KEY]
        return clip_normalize(decoded)


@LossProtocol.register("adain")
class AdaINLoss(LossProtocol):
    def _init_config(self) -> None:
        self.mse = nn.MSELoss()
        self.content_w = self.config.setdefault("content_weight", 1.0)
        self.style_w = self.config.setdefault("style_weight", 10.0)

    def _content_loss(self, stylized: Tensor, target: Tensor) -> Tensor:
        return self.mse(stylized, target)

    def _style_loss(self, stylized: Tensor, target: Tensor) -> Tensor:
        stylized_mean, stylized_std = mean_std(stylized)
        target_mean, target_std = mean_std(target)
        mean_loss = self.mse(stylized_mean, target_mean)
        std_loss = self.mse(stylized_std, target_std)
        return mean_loss + std_loss

    def _core(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> losses_type:
        content_latent = forward_results[CONTENT_LATENT_KEY]
        stylized_latent = forward_results[STYLIZED_CONTENT_LATENT_KEY]
        content_loss = self._content_loss(stylized_latent, content_latent)
        style_feats = forward_results[STYLE_LATENTS_KEY]
        stylized_feats = forward_results[STYLIZED_STYLE_LATENTS_KEY]
        style_loss = sum(
            self._style_loss(stylized, style_feats[k])
            for k, stylized in stylized_feats.items()
        )
        loss = self.content_w * content_loss + self.style_w * style_loss
        return {LOSS_KEY: loss, "content": content_loss, "style": style_loss}


__all__ = [
    "AdaINStylizer",
    "AdaINLoss",
]



import os
import math
import logging
import warnings

import numba as nb
import numpy as np

from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import Optional
from typing import NamedTuple
from cfc.stat import RollingStat
from cftool.misc import hash_code
from cftool.misc import fix_float_to_length
from cftool.misc import LoggingMixin
from cftool.misc import WithRegister
from cftool.array import squeeze
from cftool.array import StrideArray
from cftool.array import SharedArrayWrapper

from ..toolkit import to_torch
from ...data import get_weighted_indices
from ...data import MLDataset
from ...types import tensor_dict_type
from ...protocol import DataLoaderProtocol
from ...constants import CACHE_DIR
from ...constants import INPUT_KEY
from ...constants import LABEL_KEY
from ...constants import BATCH_INDICES_KEY


class DataConfig(NamedTuple):
    x_window: int
    y_window: int  # y_window could be 0, which means it will use x_window as its value
    nan_ratio: float
    gap: int
    train_split_tick: str
    train_start_tick: Optional[str] = None
    test_split_tick: Optional[str] = None
    test_end_tick: Optional[str] = None
    use_target_as_input: bool = True
    sample_by_ticks: bool = False
    nan_fill: Union[str, float] = 0.0
    callback: str = "batch_first_prod"
    callback_config: Optional[Dict[str, Any]] = None
    filter_fns: Optional[Dict[str, str]] = None
    filter_fns_config: Optional[Dict[str, Dict[str, Any]]] = None
    xy_padding: int = 1
    data_cache_folder: str = ".data_cache"
    use_memory_sharing: bool = True

    @property
    def y_span(self) -> int:
        return self.y_window + self.xy_padding


filter_fns: Dict[str, Type["FilterFunctions"]] = {}
data_callbacks: Dict[str, Type["DataCallback"]] = {}


class DataCallback(WithRegister["DataCallback"], metaclass=ABCMeta):
    d = data_callbacks

    def __init__(self, cfg: DataConfig, is_clf: bool):
        self.cfg = cfg
        self.is_clf = is_clf
        self.config = cfg.callback_config or {}

    @abstractmethod
    def process_batch(self, x: np.ndarray, labels: np.ndarray) -> tensor_dict_type:
        pass

    @staticmethod
    def _to_dict(x: np.ndarray, labels: np.ndarray) -> tensor_dict_type:
        return dict(zip([INPUT_KEY, LABEL_KEY], map(to_torch, [x, labels])))

    @classmethod
    def make_with(cls, cfg: DataConfig, is_clf: bool) -> "DataCallback":
        return data_callbacks[cfg.callback](cfg, is_clf)


@DataCallback.register("identity")
class Identity(DataCallback):
    def process_batch(self, x: np.ndarray, labels: np.ndarray) -> tensor_dict_type:
        return self._to_dict(x, labels)


class BatchFirstAggregate(DataCallback, metaclass=ABCMeta):
    def __init__(self, cfg: DataConfig, is_clf: bool):
        super().__init__(cfg, is_clf)
        self.clf_threshold = self.config.setdefault("clf_threshold", 0.0015)
        self.label_multiplier = self.config.setdefault("label_multiplier", None)

    @abstractmethod
    def aggregate_labels(self, labels: np.ndarray) -> np.ndarray:
        pass

    def process_batch(self, x: np.ndarray, labels: np.ndarray) -> tensor_dict_type:
        if labels.shape[1] > 1:
            labels = self.aggregate_labels(labels)
        if self.is_clf:
            labels = (labels > self.clf_threshold).astype(np.int64)
        elif self.label_multiplier is not None:
            labels *= self.label_multiplier
        return self._to_dict(x, labels)


@DataCallback.register("batch_first_prod")
class BatchFirstProd(BatchFirstAggregate):
    def aggregate_labels(self, labels: np.ndarray) -> np.ndarray:
        return np.prod(labels + 1.0, axis=-1, keepdims=True) - 1.0


@DataCallback.register("batch_first_cum_prod")
class BatchFirstCumProd(BatchFirstAggregate):
    def aggregate_labels(self, labels: np.ndarray) -> np.ndarray:
        return np.cumprod(labels + 1.0, axis=-1) - 1.0


class FilterFunctions(WithRegister["FilterFunctions"]):
    d = filter_fns

    def __init__(self, key: str, cfg: DataConfig, is_clf: bool):
        self.key = key
        self.cfg = cfg
        self.is_clf = is_clf
        self.config = (cfg.filter_fns_config or {}).get(key, {})

    def get_invalid_tick_mask(self, ticks: np.ndarray) -> np.ndarray:
        pass

    def get_invalid_objects_mask(self, objects: List[str]) -> np.ndarray:
        pass

    def get_valid_labels_mask(self, labels: np.ndarray) -> np.ndarray:
        pass

    @classmethod
    def make_with(cls, cfg: DataConfig, is_clf: bool) -> Dict[str, "FilterFunctions"]:
        if cfg.filter_fns is None:
            raise ValueError("`filter_fns` is not provided")
        return {k: filter_fns[v](k, cfg, is_clf) for k, v in cfg.filter_fns.items()}


@FilterFunctions.register("keep_edge")
class KeepEdge(FilterFunctions):
    def __init__(self, key: str, cfg: DataConfig, is_clf: bool):
        super().__init__(key, cfg, is_clf)
        data_callback = DataCallback.make_with(cfg, is_clf)
        assert isinstance(data_callback, BatchFirstAggregate)
        self.data_callback = data_callback

    def get_valid_labels_mask(self, labels: np.ndarray) -> np.ndarray:
        labels = np.transpose(labels, [0, 2, 1])
        ratio = self.config.setdefault("ratio", 0.1)
        aggregated = squeeze(self.data_callback.aggregate_labels(labels))
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", r"All-NaN (slice|axis) encountered")
            floor = np.nanquantile(aggregated, ratio, axis=1, keepdims=True)
            ceiling = np.nanquantile(aggregated, 1.0 - ratio, axis=1, keepdims=True)
        return (aggregated <= floor) | (aggregated >= ceiling)


@nb.jit(nopython=True, parallel=True)
def rolling_valid(x: np.ndarray, window: int) -> np.ndarray:
    nb.set_num_threads(8)
    out = np.full(x.shape[:2], True)
    for i in nb.prange(out.shape[0]):
        for j in range(out.shape[1]):
            if np.any(x[i, j]):
                out[i, j : j + window] = False
    return out[..., window - 1 :]


def pad_head(arr: np.ndarray, pad_width: int, pad_value: Any) -> np.ndarray:
    return np.pad(
        arr,
        pad_width=[[pad_width, 0], [0, 0]],
        mode="constant",
        constant_values=pad_value,
    )


def get_valid_mask(arr: np.ndarray, window: int, nan_ratio: float) -> np.ndarray:
    nan_mask = np.isnan(arr)
    if nan_ratio <= 0.0:
        all_valid = rolling_valid(nan_mask.transpose([1, 0, 2]), window).T
    else:
        nan_mask = nan_mask.astype(np.float32)
        nan_ratio_mat = RollingStat.mean(nan_mask, window, axis=0)
        all_valid = np.all(nan_ratio_mat <= nan_ratio, axis=-1)
    return pad_head(all_valid, window - 1, False)


class TSBundle(NamedTuple):
    x: np.ndarray
    y: np.ndarray
    ticks: np.ndarray
    indices: np.ndarray
    offset: int


class SpanSplit(NamedTuple):
    train_span: np.ndarray
    valid_span: np.ndarray
    test_span: Optional[np.ndarray] = None


class TimeSeriesDataManager(LoggingMixin):
    data: MLDataset

    def __init__(
        self,
        cfg: DataConfig,
        code: str,
        *,
        is_clf: bool,
        cache_folder: str = os.path.join(CACHE_DIR, "ts"),
    ):
        self.cfg = cfg
        self.code = code
        self.is_clf = is_clf
        self.cache_folder = cache_folder

        # saw
        def _get_saw(file: str) -> SharedArrayWrapper:
            to_memory = cfg.use_memory_sharing
            return SharedArrayWrapper(code_folder, file, to_memory=to_memory)

        code_folder = os.path.join(cache_folder, code)
        os.makedirs(code_folder, exist_ok=True)
        self.x_saw = _get_saw("x.npy")
        self.y_saw = _get_saw("y.npy")
        self.ticks_saw = _get_saw("ticks.npy")
        nan_ratio_str = fix_float_to_length(cfg.nan_ratio, 6)
        window_str = f"{cfg.x_window}-{cfg.y_window}"
        filter_str = f"{cfg.filter_fns}_{cfg.filter_fns_config}"
        valid_mask_code = hash_code(f"{nan_ratio_str}{window_str}{filter_str}")
        self.valid_mask_saw = _get_saw(f"valid_mask_{valid_mask_code}.npy")

    @property
    def is_ready(self) -> bool:
        if not self.x_saw.is_ready:
            return False
        if not self.y_saw.is_ready:
            return False
        if not self.valid_mask_saw.is_ready:
            return False
        return True

    def _get_span(self, ticks: np.ndarray, mask: np.ndarray, gap: int) -> np.ndarray:
        indices = np.nonzero(mask)[0]
        if gap > 0:
            indices = indices[gap:]
        if indices.shape[0] <= 0:
            raise ValueError(
                "tick settings are invalid, "
                "which results in no valid data are available"
            )
        # pad first
        first_index = indices[0]
        padded_indices = first_index - np.arange(self.cfg.x_window - 1)[::-1] - 1
        padded_indices = padded_indices[padded_indices >= 0]
        indices = np.append(padded_indices, indices)
        # pad last
        if self.cfg.y_window > 0:
            last_index = indices[-1]
            padded_indices = last_index + np.arange(self.cfg.y_span) + 1
            padded_indices = padded_indices[padded_indices < len(ticks)]
            indices = np.append(indices, padded_indices)
        return indices[[0, -1]]

    def make_data(
        self,
        x: Optional[np.ndarray],
        y: Optional[np.ndarray],
        ticks: Optional[np.ndarray],
    ) -> None:
        # sanity check
        if x is None:
            if not self.x_saw.is_ready:
                raise ValueError("`x` should be provided because it is not ready")
            x = self.x_saw.read()
        if y is None:
            if not self.y_saw.is_ready:
                raise ValueError("`y` should be provided because it is not ready")
            y = self.y_saw.read()
        if ticks is None:
            if not self.ticks_saw.is_ready:
                raise ValueError("`ticks` should be provided because it is not ready")
            ticks = self.ticks_saw.read()
        if len(x.shape) != 3:
            raise ValueError("`x` should be 3D array (ticks, objects, dim)")
        if len(x) != len(y):
            raise ValueError("length of `x` should equal to length of `y`")
        if len(x) != len(ticks):
            raise ValueError("length of `x` should equal to length of `ticks`")
        # filters
        if self.cfg.filter_fns is None:
            filters = None
        else:
            filters = FilterFunctions.make_with(self.cfg, self.is_clf)
        # valid mask
        x_window = self.cfg.x_window
        y_window = self.cfg.y_window or x_window
        x_mask = get_valid_mask(x, x_window, self.cfg.nan_ratio)
        y_mask = get_valid_mask(y[..., None], y_window, self.cfg.nan_ratio)
        if filters is not None:
            # for x & y, only train dataset needs to be filtered
            label_filter_fn = filters.get("label")
            if y_mask is not None and label_filter_fn is not None:
                train_split = self.cfg.train_split_tick
                num_train = (ticks <= train_split).sum()
                if self.cfg.y_window > 0:
                    num_train += self.cfg.y_span
                train_target = x[:num_train]
                rolled_target = StrideArray(train_target).roll(y_window, axis=0)
                filter_y_mask = label_filter_fn.get_valid_labels_mask(rolled_target)
                y_mask[y_window - 1 : num_train] &= filter_y_mask
            # for ticks & objects, all datasets need to be filtered
            tick_filter_fn = filters.get("tick")
            if tick_filter_fn is not None:
                tick_invalid_mask = tick_filter_fn.get_invalid_tick_mask(ticks)
                x_mask[tick_invalid_mask] = False
        if self.cfg.y_window == 0:
            valid_mask = x_mask & y_mask
        else:
            offset = x_window + self.cfg.y_span - 1
            offset_valid = np.zeros_like(y_mask, bool)
            offset_valid[x_window - 1 : -self.cfg.y_span] = y_mask[offset:]
            valid_mask = x_mask & offset_valid
        # sanity check
        invalid_indices = np.nonzero(~np.any(valid_mask, 1))[0][x_window:]
        if self.cfg.y_window > 0:
            invalid_indices = invalid_indices[: -self.cfg.y_span]
        if invalid_indices.shape[0] != 0 and self.cfg.filter_fns is None:
            self.log_msg(
                "invalid data occurred\n"
                f"* ticks   : {ticks[invalid_indices].tolist()}\n"
                f"* indices : {invalid_indices.tolist()}",
                self.warning_prefix,
                msg_level=logging.WARNING,
            )
        # finalize
        self.x_saw.write(x)
        self.y_saw.write(y)
        self.ticks_saw.write(ticks)
        self.valid_mask_saw.write(valid_mask)

    def split(self) -> SpanSplit:
        if not self.is_ready:
            raise ValueError("`TimeSeriesDataManager` is not ready yet")
        ticks = self.ticks_saw.read()
        # (train_start -> ) train_split -> test_split -> test_end
        if (
            self.cfg.test_split_tick is None
            or self.cfg.test_split_tick > self.cfg.train_split_tick
        ):
            train_mask = ticks <= self.cfg.train_split_tick
            remained_mask = ~train_mask
            if self.cfg.train_start_tick is not None:
                train_mask &= ticks >= self.cfg.train_start_tick
            if self.cfg.test_split_tick is None:
                validation_mask = test_mask = None
                if self.cfg.test_end_tick is not None:
                    remained_mask &= ticks <= self.cfg.test_end_tick
            else:
                test_mask = ticks > self.cfg.test_split_tick
                validation_mask = remained_mask & ~test_mask
                if self.cfg.test_end_tick is not None:
                    test_mask &= ticks <= self.cfg.test_end_tick
        # (train_start -> ) test_split -> train_split -> test_end
        else:
            validation_mask = ticks <= self.cfg.test_split_tick
            remained_mask = ~validation_mask
            if self.cfg.train_start_tick is not None:
                validation_mask &= ticks >= self.cfg.train_start_tick
            test_mask = ticks > self.cfg.train_split_tick
            train_mask = remained_mask & ~test_mask
            if self.cfg.test_end_tick is not None:
                test_mask &= ticks <= self.cfg.test_end_tick
        # train
        train_span = self._get_span(ticks, train_mask, 0)
        # valid
        if self.cfg.test_split_tick is None:
            valid_span = self._get_span(ticks, remained_mask, self.cfg.gap)
            return SpanSplit(train_span, valid_span)
        assert validation_mask is not None
        valid_span = self._get_span(ticks, validation_mask, self.cfg.gap)
        # test
        assert test_mask is not None
        test_span = self._get_span(ticks, test_mask, self.cfg.gap)
        return SpanSplit(train_span, valid_span, test_span)

    def pick_from_span(
        self,
        span: np.ndarray,
        masks: Optional[List[np.ndarray]] = None,
    ) -> TSBundle:
        # sanity check
        x, y = self.x_saw.read(), self.y_saw.read()
        if not self.is_ready:
            raise ValueError("`TimeSeriesDataManager` is not ready yet")
        if masks is not None:
            for mask in masks:
                if mask.shape != x.shape[:2]:
                    raise ValueError("`masks` is not compatible with `x`")
        # pick
        begin, end = span.tolist()
        # x & y
        x, y = x[begin:end], y[begin:end]
        # ticks
        ticks = self.ticks_saw.read()[begin:end]
        # indices
        valid_mask = self.valid_mask_saw.read().copy()
        if masks is not None:
            offset_start = self.cfg.x_window - 1
            padding = self.cfg.xy_padding + 1
            offset = offset_start + padding
            for mask in masks:
                local_valid = np.zeros_like(mask, bool)
                local_valid[offset_start:-padding] = mask[offset:]
                valid_mask &= local_valid
        valid_mask[begin : begin + self.cfg.x_window - 1] = False
        if self.cfg.y_window > 0:
            valid_mask[end - self.cfg.y_span : end] = False
        tick_indices, object_indices = np.nonzero(valid_mask)
        indices = np.vstack([tick_indices, object_indices]).T
        return TSBundle(x, y, ticks, indices, begin)


@DataLoaderProtocol.register("ts")
class TimeSeriesLoader(DataLoaderProtocol):
    data: MLDataset

    def __init__(
        self,
        cfg: DataConfig,
        bundle: TSBundle,
        *,
        is_clf: bool,
        shuffle: bool,
        batch_size: int = 128,
        sample_weights: Optional[np.ndarray] = None,
    ):
        if sample_weights is not None and len(bundle.x) != len(sample_weights):
            raise ValueError(
                f"the number of data samples ({len(bundle.x)}) is not identical with "
                f"the number of sample weights ({len(sample_weights)})"
            )
        super().__init__(sample_weights=sample_weights)
        self.name = None
        self.cfg = cfg
        self.bundle = bundle
        self.is_clf = is_clf
        self.shuffle = shuffle
        self.shuffle_backup = shuffle
        self.batch_size = batch_size
        self.callback = DataCallback.make_with(cfg, is_clf)
        self.indices = self.bundle.indices.copy() - self.bundle.offset
        # prepare data
        self.num_ticks, self.num_objects, self.dim = bundle.x.shape
        x_window = cfg.x_window
        y_window = cfg.y_window
        rolled_x = StrideArray(bundle.x).roll(x_window, axis=0)
        rolled_y = StrideArray(bundle.y).roll(y_window or x_window, axis=0)
        self.data = MLDataset(rolled_x, rolled_y)

    def __len__(self) -> int:
        return math.ceil(self.bundle.indices.shape[0] / self.batch_size)

    def __iter__(self) -> "TimeSeriesLoader":
        self._cursor = -1
        indices = self.indices.copy()
        if self.shuffle:
            weighted = get_weighted_indices(self.indices.shape[0], self.sample_weights)
            indices = self.indices[np.random.permutation(weighted)]
        self._indices = indices
        return self

    def __next__(self) -> tensor_dict_type:
        self._cursor += 1
        if self._cursor == len(self):
            raise StopIteration
        # indices
        start = self._cursor * self.batch_size
        end = start + self.batch_size
        # if the next batch is the last batch and its size is too small,
        #  merge it to the current batch
        if 0 < self.num_samples - end <= 0.5 * self.batch_size:
            end = self.num_samples
            self._cursor += 1
        ticks_indices, objects_indices = self._indices[start:end]
        x_ticks_indices = ticks_indices - (self.cfg.x_window - 1)
        # x & y
        x_batch = self.data.x[x_ticks_indices, ..., objects_indices, :]
        if self.cfg.y_window == 0:
            y_ticks_indices = x_ticks_indices
        else:
            y_ticks_indices = ticks_indices + self.cfg.xy_padding + 1
        labels = self.data.y[y_ticks_indices, ..., objects_indices]  # type: ignore
        # protocol
        batch_indices = ticks_indices * self.num_objects + objects_indices
        batch_indices = batch_indices.reshape([-1, 1])
        x_batch, labels = self._handle_nan(x_batch, labels)
        sample = self.callback.process_batch(x_batch, labels)
        sample[BATCH_INDICES_KEY] = to_torch(batch_indices)
        return sample

    def disable_shuffle(self) -> None:
        self.shuffle = False

    def recover_shuffle(self) -> None:
        self.shuffle = self.shuffle_backup

    def copy(self) -> "TimeSeriesLoader":
        return TimeSeriesLoader(
            self.cfg,
            self.bundle,
            is_clf=self.is_clf,
            shuffle=self.shuffle,
            batch_size=self.batch_size,
            sample_weights=self.sample_weights,
        )

    @property
    def num_samples(self) -> int:
        return len(self.indices)

    def _handle_nan(
        self,
        x_batch: np.ndarray,
        labels: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        if self.cfg.nan_ratio <= 0.0:
            return x_batch, labels
        nan_fill = self.cfg.nan_fill
        x_nan, y_nan = np.isnan(x_batch), np.isnan(labels)
        if isinstance(nan_fill, float):
            x_batch[x_nan] = labels[y_nan] = nan_fill
        else:
            msg = f"nan fill method '{nan_fill}' is not defined"
            raise NotImplementedError(msg)
        if self.cfg.y_window == 0:
            labels = labels[..., [-1]]
        return x_batch, labels


__all__ = [
    "DataCallback",
    "FilterFunctions",
    "TimeSeriesLoader",
]



from torch import Tensor
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Optional

from .core import Text2ImageAligner
from ..clip import CLIP
from ...cv.generator import VQGANGenerator


@Text2ImageAligner.register("clip_vqgan_aligner")
class CLIPWithVQGANAligner(Text2ImageAligner):
    perceptor: CLIP
    generator: VQGANGenerator

    def __init__(
        self,
        perceptor: str = "clip",
        generator: str = "vqgan_generator",
        tokenizer: str = "clip",
        resolution: Tuple[int, int] = (400, 224),
        *,
        text: str,
        noise: str = "fractal",
        condition_path: Optional[str] = None,
        vision_encoder: Optional[str] = None,
        vision_encoder_config: Optional[Dict[str, Any]] = None,
        vision_encoder_pretrained_name: Optional[str] = None,
        vision_encoder_pretrained_path: Optional[str] = None,
        vision_weight: float = 0.1,
        vision_monitor_step: int = 5,
        tokenizer_config: Optional[Dict[str, Any]] = None,
        num_cuts: int = 36,
        perceptor_config: Optional[Dict[str, Any]] = None,
        generator_config: Optional[Dict[str, Any]] = None,
        perceptor_pretrained_name: Optional[str] = "clip",
        generator_pretrained_name: Optional[str] = "vqgan_generator",
        perceptor_pretrained_path: Optional[str] = None,
        generator_pretrained_path: Optional[str] = None,
    ):
        if generator_config is None:
            generator_config = {}
        generator_config.setdefault("img_size", 256)
        if condition_path is not None:
            if vision_encoder is None:
                vision_encoder = "backbone"
            if vision_encoder_config is None:
                vision_encoder_config = {"name": "vgg_style", "pretrained": True}
        super().__init__(
            perceptor,
            generator,
            tokenizer,
            resolution,
            text=text,
            noise=noise,
            condition_path=condition_path,
            vision_encoder=vision_encoder,
            vision_encoder_config=vision_encoder_config,
            vision_encoder_pretrained_name=vision_encoder_pretrained_name,
            vision_encoder_pretrained_path=vision_encoder_pretrained_path,
            vision_weight=vision_weight,
            vision_monitor_step=vision_monitor_step,
            tokenizer_config=tokenizer_config,
            num_cuts=num_cuts,
            perceptor_config=perceptor_config,
            generator_config=generator_config,
            perceptor_pretrained_name=perceptor_pretrained_name,
            generator_pretrained_name=generator_pretrained_name,
            perceptor_pretrained_path=perceptor_pretrained_path,
            generator_pretrained_path=generator_pretrained_path,
        )
        self.normalize = self.perceptor.get_transform().transforms[-1]

    def generate_raw(self) -> Tensor:
        z_q = self.generator.codebook(self.z)[0]
        return self.generator.decode(z_q, resize=False)

    def perceptor_normalize(self, image: Tensor) -> Tensor:
        return self.normalize(image)


__all__ = [
    "CLIPWithVQGANAligner",
]



import os
import sys
import json
import math
import torch
import logging
import argparse
import torchvision
import urllib.request

import numpy as np
import torch.nn as nn
import matplotlib.pyplot as plt
import torch.nn.functional as F

from PIL import Image
from PIL import ImageDraw
from torch import Tensor
from typing import Any
from typing import Set
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from typing import ContextManager
from zipfile import ZipFile
from collections import defaultdict
from collections import OrderedDict
from onnxruntime import InferenceSession
from torch.optim import Optimizer
from cftool.misc import prod
from cftool.misc import check_requires
from cftool.misc import shallow_copy_dict
from cftool.misc import context_error_handler
from cftool.misc import LoggingMixin
from cftool.misc import DownloadProgressBar
from cftool.array import to_standard
from cfml.misc.toolkit import show_or_save

from ..types import arr_type
from ..types import data_type
from ..types import param_type
from ..types import np_dict_type
from ..types import tensor_dict_type
from ..types import sample_weights_type
from ..constants import CACHE_DIR
from ..constants import INPUT_KEY
from ..constants import INFO_PREFIX
from ..constants import WARNING_PREFIX


# general


def check_is_ci() -> bool:
    parser = argparse.ArgumentParser()
    parser.add_argument("--ci", type=int, default=0)
    args = parser.parse_args()
    return bool(args.ci)


def check_available(tag: str, repo: str, name: str) -> bool:
    with open(os.path.join(os.path.dirname(__file__), "available.json"), "r") as f:
        available = json.load(f)
    if repo != "pretrained-models":
        return True
    return name in available[tag]


def download(
    tag: str,
    repo: str,
    name: str,
    root: str,
    extension: str,
    force_download: bool,
    remove_zip: Optional[bool],
    extract_zip: bool,
) -> str:
    if not check_available(tag, repo, name):
        raise ValueError(f"'{name}' is currently not available at '{tag}'")
    prefix = f"https://github.com/carefree0910/{repo}/releases/download/{tag}/"
    os.makedirs(root, exist_ok=True)
    file = f"{name}.{extension}"
    path = os.path.join(root, file)
    is_zip = extension == "zip"
    zip_folder_path = os.path.join(root, name)
    exists = os.path.isdir(zip_folder_path) if is_zip else os.path.isfile(path)
    if exists and not force_download:
        return zip_folder_path if is_zip else path
    with DownloadProgressBar(unit="B", unit_scale=True, miniters=1, desc=name) as t:
        urllib.request.urlretrieve(
            f"{prefix}{file}",
            filename=path,
            reporthook=t.update_to,
        )
    if not is_zip:
        return path
    if extract_zip:
        with ZipFile(path, "r") as zip_ref:
            zip_ref.extractall(zip_folder_path)
    if remove_zip is None:
        remove_zip = extract_zip
    if remove_zip:
        if extract_zip:
            os.remove(path)
        else:
            print(
                f"{WARNING_PREFIX}zip file is not extracted, "
                f"so we will not remove it!"
            )
    return zip_folder_path


def download_data_info(
    name: str,
    *,
    root: str = os.path.join(CACHE_DIR, "data_info"),
    force_download: bool = False,
) -> str:
    return download(
        "data_info",
        "pretrained-models",
        name,
        root,
        "json",
        force_download,
        None,
        False,
    )


def download_tokenizer(
    name: str,
    *,
    root: str = os.path.join(CACHE_DIR, "tokenizers"),
    force_download: bool = False,
) -> str:
    return download(
        "tokenizers",
        "pretrained-models",
        name,
        root,
        "pkl",
        force_download,
        None,
        False,
    )


def download_model(
    name: str,
    *,
    root: str = os.path.join(CACHE_DIR, "models"),
    force_download: bool = False,
) -> str:
    return download(
        "checkpoints",
        "pretrained-models",
        name,
        root,
        "pt",
        force_download,
        None,
        False,
    )


def download_reference(
    name: str,
    *,
    root: str = os.path.join(CACHE_DIR, "reference"),
    force_download: bool = False,
) -> str:
    return download(
        "reference",
        "pretrained-models",
        name,
        root,
        "pt",
        force_download,
        None,
        False,
    )


def download_dataset(
    name: str,
    *,
    root: str = os.getcwd(),
    force_download: bool = False,
    remove_zip: Optional[bool] = None,
    extract_zip: bool = True,
) -> str:
    return download(
        "latest",
        "datasets",
        name,
        root,
        "zip",
        force_download,
        remove_zip,
        extract_zip,
    )


def get_compatible_name(
    tag: str,
    repo: str,
    name: str,
    versions: List[Tuple[int, int]],
    *,
    bc: bool = False,
) -> str:
    version_info = sys.version_info
    version = None
    if bc:
        tgt_versions = list(
            filter(
                lambda ver: version_info.major < ver[0] or version_info.minor < ver[1],
                versions,
            )
        )
        if tgt_versions is not None:
            version = max(tgt_versions)
    if not bc:
        tgt_versions = list(
            filter(
                lambda ver: version_info.major > ver[0] or version_info.minor >= ver[1],
                versions,
            )
        )
        if tgt_versions is not None:
            version = max(tgt_versions)
    if version is not None:
        compatible_name = f"{name}_{version[0]}.{version[1]}"
        if check_available(tag, repo, compatible_name):
            name = compatible_name
        else:
            print(
                f"{WARNING_PREFIX}compatible name '{compatible_name}' is not available "
                f"on the server, will use the original name ({name}) instead"
            )
    return name


class WeightsStrategy:
    def __init__(self, strategy: Optional[str]):
        self.strategy = strategy

    def __call__(self, num_train: int, num_valid: int) -> sample_weights_type:
        if self.strategy is None:
            return None
        return getattr(self, self.strategy)(num_train, num_valid)

    def linear_decay(self, num_train: int, num_valid: int) -> sample_weights_type:
        return np.linspace(0, 1, num_train + 1)[1:]

    def radius_decay(self, num_train: int, num_valid: int) -> sample_weights_type:
        return np.sin(np.arccos(1.0 - np.linspace(0, 1, num_train + 1)[1:]))

    def log_decay(self, num_train: int, num_valid: int) -> sample_weights_type:
        return np.log(np.arange(num_train) + np.e)

    def sigmoid_decay(self, num_train: int, num_valid: int) -> sample_weights_type:
        x = np.linspace(-5.0, 5.0, num_train)
        return 1.0 / (1.0 + np.exp(-x))

    def visualize(self, export_path: str = "weights_strategy.png") -> None:
        n = 1000
        x = np.linspace(0, 1, n)
        y = self(n, 0)
        if isinstance(y, tuple):
            y = y[0]
        plt.figure()
        plt.plot(x, y)
        show_or_save(export_path)


class LoggingMixinWithRank(LoggingMixin):
    is_rank_0: bool = True

    def set_rank_0(self, value: bool) -> None:
        self.is_rank_0 = value
        for v in self.__dict__.values():
            if isinstance(v, LoggingMixinWithRank):
                v.set_rank_0(value)

    def _init_logging(
        self,
        verbose_level: Optional[int] = 2,
        trigger: bool = True,
    ) -> None:
        if not self.is_rank_0:
            return None
        super()._init_logging(verbose_level, trigger)

    def log_msg(
        self,
        body: str,
        prefix: str = "",
        verbose_level: Optional[int] = 1,
        msg_level: int = logging.INFO,
        frame: Any = None,
    ) -> None:
        if not self.is_rank_0:
            return None
        super().log_msg(body, prefix, verbose_level, msg_level, frame)

    def log_block_msg(
        self,
        body: str,
        prefix: str = "",
        title: str = "",
        verbose_level: Optional[int] = 1,
        msg_level: int = logging.INFO,
        frame: Any = None,
    ) -> None:
        if not self.is_rank_0:
            return None
        super().log_block_msg(body, prefix, title, verbose_level, msg_level, frame)

    def log_timing(self) -> None:
        if not self.is_rank_0:
            return None
        return super().log_timing()


# dl


def inject_debug(config: Dict[str, Any]) -> None:
    config["fixed_steps"] = 1
    config["valid_portion"] = 1.0e-4


def fix_denormal_states(
    states: tensor_dict_type,
    *,
    eps: float = 1.0e-32,
    verbose: bool = False,
) -> tensor_dict_type:
    new_states = shallow_copy_dict(states)
    num_total = num_denormal_total = 0
    for k, v in states.items():
        if not v.is_floating_point():
            continue
        num_total += v.numel()
        denormal = (v != 0) & (v.abs() < eps)
        num_denormal = denormal.sum().item()
        num_denormal_total += num_denormal
        if num_denormal > 0:
            new_states[k][denormal] = v.new_zeros(num_denormal)
    if verbose:
        print(f"{INFO_PREFIX}denormal ratio : {num_denormal_total / num_total:8.6f}")
    return new_states


def toggle_optimizer(m: nn.Module, optimizer: Optimizer) -> None:
    for param in m.parameters():
        param.requires_grad = False
    for group in optimizer.param_groups:
        for param in group["params"]:
            param.requires_grad = True


def to_torch(arr: np.ndarray) -> Tensor:
    return torch.from_numpy(to_standard(arr))


def to_numpy(tensor: Tensor) -> np.ndarray:
    return tensor.detach().cpu().numpy()


def to_device(
    batch: tensor_dict_type,
    device: torch.device,
    **kwargs: Any,
) -> tensor_dict_type:
    return {
        k: v.to(device, **kwargs)
        if isinstance(v, Tensor)
        else [vv.to(device, **kwargs) if isinstance(vv, Tensor) else vv for vv in v]
        if isinstance(v, list)
        else v
        for k, v in batch.items()
    }


def has_batch_norms(m: nn.Module) -> bool:
    bn_types = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm)
    for name, module in m.named_modules():
        if isinstance(module, bn_types):
            return True
    return False


def inject_parameters(
    src: nn.Module,
    tgt: nn.Module,
    *,
    strict: Optional[bool] = None,
    src_filter_fn: Optional[Callable[[str], bool]] = None,
    tgt_filter_fn: Optional[Callable[[str], bool]] = None,
    custom_mappings: Optional[Dict[str, str]] = None,
    states_callback: Optional[Callable[[tensor_dict_type], tensor_dict_type]] = None,
) -> None:
    if strict is None:
        strict = tgt_filter_fn is None
    src_states = src.state_dict()
    tgt_states = tgt.state_dict()
    if src_filter_fn is not None:
        pop_keys = [key for key in src_states if not src_filter_fn(key)]
        for key in pop_keys:
            src_states.pop(key)
    if tgt_filter_fn is not None:
        pop_keys = [key for key in tgt_states if not tgt_filter_fn(key)]
        for key in pop_keys:
            tgt_states.pop(key)
    if states_callback is not None:
        src_states = states_callback(shallow_copy_dict(src_states))
    if len(src_states) != len(tgt_states):
        raise ValueError(f"lengths of states are not identical between {src} and {tgt}")
    new_states = OrderedDict()
    if custom_mappings is not None:
        for src_k, tgt_k in custom_mappings.items():
            new_states[tgt_k] = src_states.pop(src_k)
            tgt_states.pop(tgt_k)
    for (src_k, src_v), (tgt_k, tgt_v) in zip(src_states.items(), tgt_states.items()):
        if src_v.shape != tgt_v.shape:
            raise ValueError(
                f"shape of {src_k} ({list(src_v.shape)}) is not identical with "
                f"shape of {tgt_k} ({list(tgt_v.shape)})"
            )
        new_states[tgt_k] = src_v
    tgt.load_state_dict(new_states, strict=strict)


class Diffs(NamedTuple):
    names1: List[str]
    names2: List[str]
    diffs: List[Tensor]


def sorted_param_diffs(m1: nn.Module, m2: nn.Module) -> Diffs:
    names1, params1 = list(zip(*m1.named_parameters()))
    names2, params2 = list(zip(*m2.named_parameters()))
    if len(params1) != len(params2):
        raise ValueError(f"lengths of params are not identical between {m1} and {m2}")
    diffs = []
    for p1, p2 in zip(params1, params2):
        (p1, _), (p2, _) = map(torch.sort, [p1.view(-1), p2.view(-1)])
        diffs.append(torch.abs(p1.data - p2.data))
    return Diffs(names1, names2, diffs)


def get_gradient(
    y: Tensor,
    x: Tensor,
    retain_graph: bool = False,
    create_graph: bool = False,
) -> Union[Tensor, Tuple[Tensor, ...]]:
    grads = torch.autograd.grad(y, x, torch.ones_like(y), retain_graph, create_graph)
    if len(grads) == 1:
        return grads[0]
    return grads


def set_requires_grad(module: nn.Module, requires_grad: bool = False) -> None:
    for param in module.parameters():
        param.requires_grad = requires_grad


def scheduler_requires_metric(scheduler: Any) -> bool:
    return check_requires(scheduler.step, "metrics")


# This is a modified version of https://github.com/sksq96/pytorch-summary
#  So it can summary `carefree-learn` model structures better
def summary(
    model: nn.Module,
    sample_batch: tensor_dict_type,
    *,
    return_only: bool = False,
) -> str:
    def _get_param_counts(module_: nn.Module) -> Tuple[int, int]:
        num_params = 0
        num_trainable_params = 0
        for param in module_.parameters():
            local_num_params = int(round(prod(param.data.shape)))
            num_params += local_num_params
            if param.requires_grad:
                num_trainable_params += local_num_params
        return num_params, num_trainable_params

    def register_hook(module: nn.Module) -> None:
        def hook(module_: nn.Module, inp: Any, output: Any) -> None:
            m_name = module_names.get(module_)
            if m_name is None:
                return

            inp = inp[0]
            if not isinstance(inp, Tensor):
                return
            if isinstance(output, (list, tuple)):
                for element in output:
                    if not isinstance(element, Tensor):
                        return
            elif not isinstance(output, Tensor):
                return

            m_dict: OrderedDict[str, Any] = OrderedDict()
            m_dict["input_shape"] = list(inp.size())
            if len(m_dict["input_shape"]) > 0:
                m_dict["input_shape"][0] = -1
            if isinstance(output, (list, tuple)):
                m_dict["output_shape"] = [[-1] + list(o.size())[1:] for o in output]
                m_dict["is_multiple_output"] = True
            else:
                m_dict["output_shape"] = list(output.size())
                if len(m_dict["output_shape"]) > 0:
                    m_dict["output_shape"][0] = -1
                m_dict["is_multiple_output"] = False

            num_params_, num_trainable_params_ = _get_param_counts(module_)
            m_dict["num_params"] = num_params_
            m_dict["num_trainable_params"] = num_trainable_params_
            raw_summary_dict[m_name] = m_dict

        if not isinstance(module, torch.jit.ScriptModule):
            hooks.append(module.register_forward_hook(hook))

    # get names
    def _inject_names(m: nn.Module, previous_names: List[str]) -> None:
        info_list = []
        for child in m.children():
            current_names = previous_names + [type(child).__name__]
            current_name = ".".join(current_names)
            module_names[child] = current_name
            info_list.append((child, current_name, current_names))
        counts: Dict[str, int] = defaultdict(int)
        idx_mapping: Dict[nn.Module, int] = {}
        for child, current_name, _ in info_list:
            idx_mapping[child] = counts[current_name]
            counts[current_name] += 1
        for child, current_name, current_names in info_list:
            if counts[current_name] == 1:
                continue
            current_name = f"{current_name}-{idx_mapping[child]}"
            module_names[child] = current_name
            current_names[-1] = current_name.split(".")[-1]
        for child, _, current_names in info_list:
            _inject_names(child, current_names)

    module_names: OrderedDict[nn.Module, str] = OrderedDict()
    existing_names: Set[str] = set()

    def _get_name(original: str) -> str:
        count = 0
        final_name = original
        while final_name in existing_names:
            count += 1
            final_name = f"{original}_{count}"
        existing_names.add(final_name)
        return final_name

    model_name = _get_name(type(model).__name__)
    module_names[model] = model_name
    _inject_names(model, [model_name])

    # create properties
    raw_summary_dict: OrderedDict[str, Any] = OrderedDict()
    hooks: List[Any] = []

    # register hook
    model.apply(register_hook)

    # make a forward pass
    with eval_context(model, use_grad=None):
        if not hasattr(model, "summary_forward"):
            model(0, sample_batch)
        else:
            model.summary_forward(0, sample_batch)  # type: ignore
        for param in model.parameters():
            param.grad = None

    # remove these hooks
    for h in hooks:
        h.remove()

    # get hierarchy
    hierarchy: OrderedDict[str, Any] = OrderedDict()
    for key in raw_summary_dict:
        split = key.split(".")
        d = hierarchy
        for elem in split[:-1]:
            d = d.setdefault(elem, OrderedDict())
        d.setdefault(split[-1], None)

    # reconstruct summary_dict
    def _inject_summary(current_hierarchy: Any, previous_keys: List[str]) -> None:
        if previous_keys and not previous_keys[-1]:
            previous_keys.pop()
        current_layer = len(previous_keys)
        current_count = hierarchy_counts.get(current_layer, 0)
        prefix = "  " * current_layer
        for k, v in current_hierarchy.items():
            current_keys = previous_keys + [k]
            concat_k = ".".join(current_keys)
            current_summary = raw_summary_dict.get(concat_k)
            summary_dict[f"{prefix}{k}-{current_count}"] = current_summary
            hierarchy_counts[current_layer] = current_count + 1
            if v is not None:
                _inject_summary(v, current_keys)

    hierarchy_counts: Dict[int, int] = {}
    summary_dict: OrderedDict[str, Any] = OrderedDict()
    _inject_summary(hierarchy, [])

    line_length = 120
    messages = ["=" * line_length]
    line_format = "{:30}  {:>20} {:>40} {:>20}"
    headers = "Layer (type)", "Input Shape", "Output Shape", "Trainable Param #"
    line_new = line_format.format(*headers)
    messages.append(line_new)
    messages.append("-" * line_length)
    total_output = 0
    for layer, layer_summary in summary_dict.items():
        layer_name = "-".join(layer.split("-")[:-1])
        if layer_summary is None:
            line_new = line_format.format(layer_name, "", "", "")
        else:
            line_new = line_format.format(
                layer_name,
                str(layer_summary["input_shape"]),
                str(layer_summary["output_shape"]),
                "{0:,}".format(layer_summary["num_trainable_params"]),
            )
            output_shape = layer_summary["output_shape"]
            is_multiple_output = layer_summary["is_multiple_output"]
            if not is_multiple_output:
                output_shape = [output_shape]
            for shape in output_shape:
                total_output += prod(shape)
        messages.append(line_new)

    total_params, trainable_params = _get_param_counts(model)
    # assume 4 bytes/number (float on cuda).
    x_batch = sample_batch[INPUT_KEY]
    get_size = lambda t: abs(prod(t.shape[1:]) * 4.0 / (1024**2.0))
    if not isinstance(x_batch, list):
        x_batch = [x_batch]
    total_input_size = sum(map(get_size, x_batch))
    # x2 for gradients
    total_output_size = abs(2.0 * total_output * 4.0 / (1024**2.0))
    total_params_size = abs(total_params * 4.0 / (1024**2.0))
    total_size = total_params_size + total_output_size + total_input_size

    non_trainable_params = total_params - trainable_params
    messages.append("=" * line_length)
    messages.append("Total params: {0:,}".format(total_params))
    messages.append("Trainable params: {0:,}".format(trainable_params))
    messages.append("Non-trainable params: {0:,}".format(non_trainable_params))
    messages.append("-" * line_length)
    messages.append("Input size (MB): %0.2f" % total_input_size)
    messages.append("Forward/backward pass size (MB): %0.2f" % total_output_size)
    messages.append("Params size (MB): %0.2f" % total_params_size)
    messages.append("Estimated Total Size (MB): %0.2f" % total_size)
    messages.append("-" * line_length)
    msg = "\n".join(messages)
    if not return_only:
        print(msg)
    return msg


class DDPInfo(NamedTuple):
    rank: int
    world_size: int
    local_rank: int


def get_ddp_info() -> Optional[DDPInfo]:
    if "RANK" in os.environ and "WORLD_SIZE" in os.environ:
        rank = int(os.environ["RANK"])
        world_size = int(os.environ["WORLD_SIZE"])
        local_rank = int(os.environ["LOCAL_RANK"])
        return DDPInfo(rank, world_size, local_rank)
    return None


def get_world_size() -> int:
    ddp_info = get_ddp_info()
    return 1 if ddp_info is None else ddp_info.world_size


class mode_context(context_error_handler):
    """
    Help entering specific mode and recovering previous mode

    This is a context controller for entering specific mode at the beginning
    and back to previous mode at the end.

    Parameters
    ----------
    module : nn.Module, arbitrary PyTorch module.

    Examples
    --------
    >>> module = nn.Module()
    >>> with mode_context(module):
    >>>     pass  # do something

    """

    def __init__(
        self,
        module: nn.Module,
        *,
        to_train: Optional[bool],
        use_grad: Optional[bool],
        use_inference: Optional[bool] = None,
    ):
        self._to_train = to_train
        self._module, self._training = module, module.training
        self._cache = {p: p.requires_grad for p in module.parameters()}
        if use_grad is not None:
            for p in module.parameters():
                p.requires_grad_(use_grad)
        if use_grad is None:
            self._grad_context: Optional[ContextManager] = None
        else:
            self._grad_context = torch.enable_grad() if use_grad else torch.no_grad()
        if use_inference is None:
            self._inference_context: Optional[ContextManager] = None
        else:
            self._inference_context = torch.inference_mode(use_inference)

    def __enter__(self) -> None:
        if self._to_train is not None:
            self._module.train(mode=self._to_train)
        if self._grad_context is not None:
            self._grad_context.__enter__()
        if self._inference_context is not None:
            self._inference_context.__enter__()

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self._to_train is not None:
            self._module.train(mode=self._training)
        if self._inference_context is not None:
            self._inference_context.__exit__(exc_type, exc_val, exc_tb)
        if self._grad_context is not None:
            self._grad_context.__exit__(exc_type, exc_val, exc_tb)
        for p, v in self._cache.items():
            p.requires_grad_(v)


class train_context(mode_context):
    """
    Useful when we need to get gradients with our PyTorch model during evaluating.
    """

    def __init__(self, module: nn.Module, *, use_grad: bool = True):
        super().__init__(module, to_train=True, use_grad=use_grad, use_inference=False)


class eval_context(mode_context):
    """
    Useful when we need to predict something with our PyTorch model during training.
    """

    def __init__(
        self,
        module: nn.Module,
        *,
        use_grad: Optional[bool] = False,
        use_inference: Optional[bool] = None,
    ):
        if use_inference is None and use_grad is not None:
            use_inference = not use_grad
        super().__init__(
            module,
            to_train=False,
            use_grad=use_grad,
            use_inference=use_inference,
        )


class Initializer(LoggingMixinWithRank):
    """
    Initializer for neural network weights

    Examples
    --------
    >>> initializer = Initializer()
    >>> linear = nn.Linear(10, 10)
    >>> initializer.xavier_uniform(linear.weight)

    """

    defined_initialization = {
        "xavier_uniform",
        "xavier_normal",
        "normal",
        "truncated_normal",
    }
    custom_initializer: Dict[str, Callable] = {}

    def __init__(self, config: Optional[Dict[str, Any]] = None):
        self.config = config or {}
        self._verbose_level = self.config.setdefault("verbose_level", 2)

    def initialize(self, param: param_type, method: str) -> Any:
        custom_initializer = self.custom_initializer.get(method)
        if custom_initializer is None:
            return getattr(self, method)(param)
        return custom_initializer(self, param)

    @classmethod
    def add_initializer(cls, f: Callable, name: str) -> None:
        if name in cls.defined_initialization:
            print(f"{WARNING_PREFIX}'{name}' initializer is already defined")
            return
        cls.defined_initialization.add(name)
        cls.custom_initializer[name] = f

    def xavier_uniform(self, param: param_type) -> None:
        gain = self.config.setdefault("gain", 1.0)
        nn.init.xavier_uniform_(param.data, gain)

    def xavier_normal(self, param: param_type) -> None:
        gain = self.config.setdefault("gain", 1.0)
        nn.init.xavier_normal_(param.data, gain)

    def normal(self, param: param_type) -> None:
        mean = self.config.setdefault("mean", 0.0)
        std = self.config.setdefault("std", 1.0)
        with torch.no_grad():
            param.data.normal_(mean, std)

    def truncated_normal(self, param: param_type) -> None:
        span = self.config.setdefault("span", 2.0)
        mean = self.config.setdefault("mean", 0.0)
        std = self.config.setdefault("std", 1.0)
        tol = self.config.setdefault("tol", 0.0)
        epoch = self.config.setdefault("epoch", 20)
        num_elem = param.numel()
        weight_base = param.new_empty(num_elem).normal_()
        get_invalid = lambda w: (w > span) | (w < -span)
        invalid = get_invalid(weight_base)
        success = False
        for _ in range(epoch):
            num_invalid = invalid.sum().item()
            if num_invalid / num_elem <= tol:
                success = True
                break
            with torch.no_grad():
                weight_base[invalid] = param.new_empty(num_invalid).normal_()
                invalid = get_invalid(weight_base)
        if not success:
            self.log_msg(
                f"invalid ratio for truncated normal : {invalid.to(torch.float32).mean():8.6f}, "
                f"it might cause by too little epoch ({epoch}) or too small tolerance ({tol})",
                prefix=self.warning_prefix,
                verbose_level=2,
                msg_level=logging.WARNING,
            )
        with torch.no_grad():
            param.data.copy_(weight_base.reshape(param.shape))
            param.data.mul_(std).add_(mean)

    def orthogonal(self, param: param_type) -> None:
        gain = self.config.setdefault("gain", 1.0)
        nn.init.orthogonal_(param.data, gain)


class DropNoGradStatesMixin:
    def state_dict(
        self,
        *,
        destination: Any = None,
        prefix: str = "",
        keep_vars: bool = False,
    ) -> tensor_dict_type:
        states = super().state_dict(destination=destination, prefix=prefix, keep_vars=keep_vars)  # type: ignore
        for key, _ in self.named_buffers():  # type: ignore
            states.pop(key)
        for key, value in self.named_parameters():  # type: ignore
            if not value.requires_grad:
                states.pop(key)
        return states

    def load_state_dict(
        self,
        state_dict: tensor_dict_type,
        strict: bool = True,
    ) -> None:
        with torch.no_grad():
            for key, value in self.named_parameters():  # type: ignore
                if value.requires_grad:
                    loaded_value = state_dict.get(key)
                    if strict and loaded_value is None:
                        raise ValueError(f"value for '{key}' is missing")
                    value.data.copy_(loaded_value)


class ONNX:
    def __init__(self, onnx_path: str):
        self.ort_session = InferenceSession(onnx_path)
        self.output_names = [node.name for node in self.ort_session.get_outputs()]

    def predict(self, new_inputs: np_dict_type) -> np_dict_type:
        if self.ort_session is None:
            raise ValueError("`onnx_path` is not provided")
        ort_inputs = {
            node.name: to_standard(new_inputs[node.name])
            for node in self.ort_session.get_inputs()
        }
        return dict(zip(self.output_names, self.ort_session.run(None, ort_inputs)))


# ml


def to_2d(arr: data_type) -> data_type:
    if arr is None or isinstance(arr, str):
        return None
    if isinstance(arr, np.ndarray):
        return arr.reshape([len(arr), -1])
    if isinstance(arr[0], list):
        return arr
    return [[elem] for elem in arr]  # type: ignore


# cv


def auto_num_layers(
    img_size: int,
    min_size: int = 4,
    target_layers: Optional[int] = 4,
    *,
    use_stride: bool = False,
) -> int:
    fn = math.ceil if use_stride else math.floor
    max_layers = fn(math.log2(img_size / min_size))
    if target_layers is None:
        return max_layers
    return max(2, min(target_layers, max_layers))


def slerp(
    x1: torch.Tensor,
    x2: torch.Tensor,
    r1: Union[float, torch.Tensor],
    r2: Optional[Union[float, torch.Tensor]] = None,
) -> torch.Tensor:
    low_norm = x1 / torch.norm(x1, dim=1, keepdim=True)
    high_norm = x2 / torch.norm(x2, dim=1, keepdim=True)
    omega = torch.acos((low_norm * high_norm).sum(1))
    so = torch.sin(omega)
    if r2 is None:
        r2 = 1.0 - r1
    x1_part = (torch.sin(r1 * omega) / so).unsqueeze(1) * x1
    x2_part = (torch.sin(r2 * omega) / so).unsqueeze(1) * x2
    return x1_part + x2_part


def interpolate(
    src: Tensor,
    *,
    mode: str = "nearest",
    factor: Optional[Union[float, Tuple[float, float]]] = None,
    size: Optional[Union[int, Tuple[int, int]]] = None,
    anchor: Optional[Tensor] = None,
    determinate: bool = False,
    **kwargs: Any,
) -> Tensor:
    if "linear" in mode or mode == "bicubic":
        kwargs.setdefault("align_corners", False)
    c, h, w = src.shape[1:]
    if determinate:
        c, h, w = map(int, [c, h, w])
    if factor is not None:
        template = "`{}` will take no affect because `factor` is provided"
        if size is not None:
            print(f"{WARNING_PREFIX}{template.format('size')}")
        if anchor is not None:
            print(f"{WARNING_PREFIX}{template.format('anchor')}")
        if factor == 1.0 or factor == (1.0, 1.0):
            return src
        if not determinate:
            return F.interpolate(
                src,
                mode=mode,
                scale_factor=factor,
                recompute_scale_factor=True,
                **kwargs,
            )
        if not isinstance(factor, tuple):
            factor = factor, factor
        size = tuple(map(int, map(round, [h * factor[0], w * factor[1]])))  # type: ignore
    if size is None:
        if anchor is None:
            raise ValueError("either `size` or `anchor` should be provided")
        size = anchor.shape[2:]
        if determinate:
            size = tuple(map(int, size))  # type: ignore
    if not isinstance(size, tuple):
        size = size, size
    if h == size[0] and w == size[1]:
        return src
    net = F.interpolate(src, size=size, mode=mode, **kwargs)
    if not determinate:
        return net
    return net.view(-1, c, *size)


def mean_std(
    latent_map: Tensor,
    eps: float = 1.0e-5,
    *,
    determinate: bool = False,
) -> Tuple[Tensor, Tensor]:
    c, h, w = latent_map.shape[1:]
    if determinate:
        c, h, w = map(int, [c, h, w])
    spatial_dim = h * w
    latent_var = latent_map.view(-1, c, spatial_dim).var(dim=2) + eps
    latent_std = latent_var.sqrt().view(-1, c, 1, 1)
    latent_mean = latent_map.view(-1, c, spatial_dim).mean(dim=2).view(-1, c, 1, 1)
    return latent_mean, latent_std


def adain_with_params(
    src: Tensor,
    mean: Tensor,
    std: Tensor,
    *,
    determinate: bool = False,
) -> Tensor:
    src_mean, src_std = mean_std(src, determinate=determinate)
    src_normalized = (src - src_mean) / src_std
    return src_normalized * std + mean


def adain_with_tensor(src: Tensor, tgt: Tensor, *, determinate: bool = False) -> Tensor:
    tgt_mean, tgt_std = mean_std(tgt, determinate=determinate)
    return adain_with_params(src, tgt_mean, tgt_std, determinate=determinate)


def make_grid(arr: arr_type, n_row: Optional[int] = None) -> Tensor:
    if isinstance(arr, np.ndarray):
        arr = to_torch(arr)
    if n_row is None:
        n_row = math.ceil(math.sqrt(len(arr)))
    return torchvision.utils.make_grid(arr, n_row)


def save_images(arr: arr_type, path: str, n_row: Optional[int] = None) -> None:
    if isinstance(arr, np.ndarray):
        arr = to_torch(arr)
    if n_row is None:
        n_row = math.ceil(math.sqrt(len(arr)))
    torchvision.utils.save_image(arr, path, normalize=True, nrow=n_row)


def make_indices_visualization_map(indices: Tensor) -> Tensor:
    images = []
    for idx in indices.view(-1).tolist():
        img = Image.new("RGB", (28, 28), (250, 250, 250))
        draw = ImageDraw.Draw(img)
        draw.text((12, 9), str(idx), (0, 0, 0))
        images.append(to_torch(np.array(img).transpose([2, 0, 1])))
    return torch.stack(images).float()



import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from .protocol import VanillaGANMixin
from ...implicit.siren import ImgSiren


@VanillaGANMixin.register("siren_gan")
class SirenGAN(VanillaGANMixin):
    def __init__(
        self,
        img_size: int,
        in_channels: int,
        out_channels: Optional[int] = None,
        latent_dim: int = 128,
        num_classes: Optional[int] = None,
        conditional_dim: int = 16,
        *,
        num_layers: int = 4,
        w_sin: float = 1.0,
        w_sin_initial: float = 30.0,
        bias: bool = True,
        final_activation: Optional[str] = None,
        discriminator: str = "basic",
        discriminator_config: Optional[Dict[str, Any]] = None,
        gan_mode: str = "vanilla",
        gan_loss_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            in_channels,
            discriminator=discriminator,
            discriminator_config=discriminator_config,
            num_classes=num_classes,
            gan_mode=gan_mode,
            gan_loss_config=gan_loss_config,
        )
        self.latent_dim = latent_dim
        self.out_channels = out_channels or in_channels
        # siren
        self.siren = ImgSiren(
            img_size,
            self.out_channels,
            latent_dim,
            num_classes,
            conditional_dim,
            num_layers=num_layers,
            w_sin=w_sin,
            w_sin_initial=w_sin_initial,
            bias=bias,
            final_activation=final_activation,
        )

    def decode(
        self,
        z: Tensor,
        *,
        labels: Optional[Tensor],
        size: Optional[int] = None,
        **kwargs: Any,
    ) -> Tensor:
        return self.siren.decode(z, labels=labels, size=size)

    # training part

    @property
    def g_parameters(self) -> List[nn.Parameter]:
        return list(self.siren.parameters())


__all__ = ["SirenGAN"]



import torch

import numpy as np
import torch.nn.functional as F

from typing import Any
from typing import List
from typing import Iterator
from typing import Optional
from sklearn.tree import _tree
from sklearn.tree import DecisionTreeClassifier

from .protocol import MERGED_KEY
from .protocol import MLCoreProtocol
from ...types import tensor_dict_type
from ...protocol import TrainerState
from ...constants import PREDICTIONS_KEY
from ...misc.toolkit import to_torch
from ...misc.toolkit import to_numpy
from ...modules.blocks import Linear
from ...modules.blocks import Activation


def export_structure(tree: DecisionTreeClassifier) -> tuple:
    tree = tree.tree_

    def recurse(node: int, depth: int) -> Iterator:
        feature_dim = tree.feature[node]
        if feature_dim == _tree.TREE_UNDEFINED:
            yield depth, -1, tree.value[node]
        else:
            threshold = tree.threshold[node]
            yield depth, feature_dim, threshold
            yield from recurse(tree.children_left[node], depth + 1)
            yield depth, feature_dim, threshold
            yield from recurse(tree.children_right[node], depth + 1)

    return tuple(recurse(0, 0))


@MLCoreProtocol.register("ndt")
class NDT(MLCoreProtocol):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        num_history: int,
        *,
        dt: DecisionTreeClassifier,
    ):
        super().__init__(in_dim, out_dim, num_history)
        tree_structure = export_structure(dt)
        # dt statistics
        num_leaves = sum([1 if pair[1] == -1 else 0 for pair in tree_structure])
        num_internals = num_leaves - 1
        # transform
        b = np.zeros(num_internals, dtype=np.float32)
        w1 = np.zeros([in_dim, num_internals], dtype=np.float32)
        w2 = np.zeros([num_internals, num_leaves], dtype=np.float32)
        w3 = np.zeros([num_leaves, out_dim], dtype=np.float32)
        node_list: List[int] = []
        node_sign_list: List[int] = []
        node_id_cursor = leaf_id_cursor = 0
        for depth, feat_dim, rs in tree_structure:
            if feat_dim != -1:
                if depth == len(node_list):
                    node_sign_list.append(-1)
                    node_list.append(node_id_cursor)
                    w1[feat_dim, node_id_cursor] = 1
                    b[node_id_cursor] = -rs
                    node_id_cursor += 1
                else:
                    node_list = node_list[: depth + 1]
                    node_sign_list = node_sign_list[:depth] + [1]
            else:
                for node_id, node_sign in zip(node_list, node_sign_list):
                    w2[node_id, leaf_id_cursor] = node_sign / len(node_list)
                w3[leaf_id_cursor] = rs / np.sum(rs)
                leaf_id_cursor += 1
        w1_, w2_, w3_, b_ = map(to_torch, [w1, w2, w3, b])
        # construct planes & routes
        self.to_planes = Linear(in_dim, num_internals, init_method=None)
        self.to_routes = Linear(num_internals, num_leaves, bias=False, init_method=None)
        self.to_leaves = Linear(num_leaves, out_dim, init_method=None)
        with torch.no_grad():
            self.to_planes.linear.bias.data = b_
            self.to_planes.linear.weight.data = w1_.t()
            self.to_routes.linear.weight.data = w2_.t()
            self.to_leaves.linear.weight.data = w3_.t()
            uniform = F.log_softmax(torch.zeros(out_dim, dtype=torch.float32), dim=0)
            self.to_leaves.linear.bias.data = uniform
        # activations
        self.planes_activation = Activation.make("sign")
        self.routes_activation = Activation.make("one_hot")

    @property
    def hyperplane_weights(self) -> np.ndarray:
        return to_numpy(self.to_planes.linear.weight)

    @property
    def hyperplane_thresholds(self) -> np.ndarray:
        return to_numpy(-self.to_planes.linear.bias)

    @property
    def route_weights(self) -> np.ndarray:
        return to_numpy(self.to_routes.linear.weight)

    @property
    def class_log_distributions(self) -> np.ndarray:
        return to_numpy(self.to_leaves.linear.weight)

    @property
    def class_log_prior(self) -> np.ndarray:
        return to_numpy(self.to_leaves.linear.bias)

    @property
    def class_prior(self) -> np.ndarray:
        return np.exp(self.class_log_prior)

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        net = batch[MERGED_KEY]
        if len(net.shape) > 2:
            net = net.contiguous().view(len(net), -1)
        net = self.planes_activation(self.to_planes(net))
        net = self.routes_activation(self.to_routes(net))
        net = self.to_leaves(net)
        return {PREDICTIONS_KEY: net}


__all__ = ["NDT"]



from torch import Tensor
from typing import List
from typing import Tuple


class IImageExtractor:
    def get_folder_latent(
        self,
        src_folder: str,
        *,
        batch_size: int,
        num_workers: int = 0,
        use_tqdm: bool = True,
    ) -> Tuple[Tensor, List[str]]:
        pass


__all__ = [
    "IImageExtractor",
]



import os
import dill
import json
import lmdb
import torch
import random
import shutil
import tempfile

import numpy as np

from abc import ABCMeta
from PIL import Image
from tqdm import tqdm
from torch import Tensor
from typing import Any
from typing import Set
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from cftool.dist import Parallel
from cftool.misc import walk
from cftool.misc import shallow_copy_dict
from cftool.misc import Saving
from torch.utils.data import Dataset
from cfdata.tabular.api import TabularData

from .core import default_lmdb_path
from .core import CVLoader
from .core import DLLoader
from .core import LMDBItem
from .core import MLLoader
from .core import CVDataset
from .core import DLDataset
from .core import MLDataset
from .core import DataLoader
from .core import Transforms
from .core import DLDataModule
from .core import ImageFolderDataset
from .core import InferenceImageFolderDataset
from ..types import data_type
from ..types import tensor_dict_type
from ..types import sample_weights_type
from ..constants import INPUT_KEY
from ..constants import LABEL_KEY
from ..constants import INFO_PREFIX
from ..constants import ERROR_PREFIX
from ..constants import WARNING_PREFIX
from ..constants import ORIGINAL_LABEL_KEY


# dl


class TensorDataset(Dataset):
    def __init__(
        self,
        x: Tensor,
        y: Optional[Tensor],
        others: Optional[tensor_dict_type] = None,
    ):
        self.x = x
        self.y = y
        self.others = others

    def __getitem__(self, index: int) -> tensor_dict_type:
        item = {INPUT_KEY: self.x[index]}
        if self.y is not None:
            label = self.y[index]
            item.update({LABEL_KEY: label, ORIGINAL_LABEL_KEY: label})
        if self.others is not None:
            for k, v in self.others.items():
                item[k] = v[index]
        return item

    def __len__(self) -> int:
        return self.x.shape[0]


class TensorDictDataset(Dataset):
    def __init__(
        self,
        x: tensor_dict_type,
        y: Optional[Tensor],
    ):
        self.x = x
        self.y = y
        self.x_keys = sorted(self.x)

    def __getitem__(self, index: int) -> tensor_dict_type:
        item = {k: self.x[k][index] for k in self.x_keys}
        if self.y is not None:
            label = self.y[index]
            item.update({LABEL_KEY: label, ORIGINAL_LABEL_KEY: label})
        return item

    def __len__(self) -> int:
        return self.x[self.x_keys[0]].shape[0]


@DLDataModule.register("tensor")
class TensorData(DLDataModule):
    def __init__(
        self,
        x_train: Tensor,
        y_train: Optional[Tensor] = None,
        x_valid: Optional[Tensor] = None,
        y_valid: Optional[Tensor] = None,
        train_others: Optional[tensor_dict_type] = None,
        valid_others: Optional[tensor_dict_type] = None,
        *,
        shuffle: bool = True,
        batch_size: int = 64,
        num_workers: int = 0,
    ):
        self.x_train = x_train
        self.y_train = y_train
        self.x_valid = x_valid
        self.y_valid = y_valid
        self.train_others = train_others
        self.valid_others = valid_others
        self.kw = dict(batch_size=batch_size, shuffle=shuffle, num_workers=num_workers)
        self.test_transform = None

    @property
    def info(self) -> Dict[str, Any]:
        return self.kw

    # TODO : support sample weights
    def prepare(self, sample_weights: sample_weights_type) -> None:
        def _get_data(x: Any, y: Any, others: Any) -> DLDataset:
            return DLDataset(TensorDataset(x, y, others))

        self.train_data = _get_data(self.x_train, self.y_train, self.train_others)
        if self.x_valid is None:
            self.valid_data = None
        else:
            self.valid_data = _get_data(self.x_valid, self.y_valid, self.valid_others)

    def initialize(self) -> Tuple[DLLoader, Optional[DLLoader]]:
        train_loader = DLLoader(DataLoader(self.train_data, **self.kw))  # type: ignore
        if self.valid_data is None:
            valid_loader = None
        else:
            valid_loader = DLLoader(DataLoader(self.valid_data, **self.kw))  # type: ignore
        return train_loader, valid_loader


@DLDataModule.register("tensor_dict")
class TensorDictData(DLDataModule):
    def __init__(
        self,
        x_train: tensor_dict_type,
        y_train: Optional[Tensor] = None,
        x_valid: Optional[tensor_dict_type] = None,
        y_valid: Optional[Tensor] = None,
        *,
        shuffle: bool = True,
        batch_size: int = 64,
        num_workers: int = 0,
    ):
        self.x_train = x_train
        self.y_train = y_train
        self.x_valid = x_valid
        self.y_valid = y_valid
        self.kw = dict(batch_size=batch_size, shuffle=shuffle, num_workers=num_workers)
        self.test_transform = None

    @property
    def info(self) -> Dict[str, Any]:
        return self.kw

    # TODO : support sample weights
    def prepare(self, sample_weights: sample_weights_type) -> None:
        def _get_data(x: Any, y: Any) -> DLDataset:
            return DLDataset(TensorDictDataset(x, y))

        self.train_data = _get_data(self.x_train, self.y_train)
        if self.x_valid is None:
            self.valid_data = None
        else:
            self.valid_data = _get_data(self.x_valid, self.y_valid)

    def initialize(self) -> Tuple[DLLoader, Optional[DLLoader]]:
        train_loader = DLLoader(DataLoader(self.train_data, **self.kw))  # type: ignore
        if self.valid_data is None:
            valid_loader = None
        else:
            valid_loader = DLLoader(DataLoader(self.valid_data, **self.kw))  # type: ignore
        return train_loader, valid_loader


@DLDataModule.register("dummy")
class DummyData(TensorData):
    def __init__(
        self,
        *,
        num_samples: int = 1,
        batch_size: int = 1,
        use_valid: bool = False,
    ):
        dummy = torch.zeros([num_samples, 1])
        x_valid = y_valid = dummy if use_valid else None
        super().__init__(dummy, dummy, x_valid, y_valid, batch_size=batch_size)


# ml


split_sw_type = Tuple[Optional[np.ndarray], Optional[np.ndarray]]


def _norm_sw(sample_weights: Optional[np.ndarray]) -> Optional[np.ndarray]:
    if sample_weights is None:
        return None
    return sample_weights / sample_weights.sum()


def _split_sw(sample_weights: sample_weights_type) -> split_sw_type:
    if sample_weights is None:
        train_weights = valid_weights = None
    else:
        if not isinstance(sample_weights, np.ndarray):
            train_weights, valid_weights = sample_weights
        else:
            train_weights, valid_weights = sample_weights, None
    train_weights, valid_weights = map(_norm_sw, [train_weights, valid_weights])
    return train_weights, valid_weights


@DLDataModule.register("ml")
class MLData(DLDataModule):
    train_data: MLDataset
    valid_data: Optional[MLDataset]

    tmp_cf_data_name = ".tmp_cf_data"

    def __init__(
        self,
        x_train: data_type,
        y_train: data_type = None,
        x_valid: data_type = None,
        y_valid: data_type = None,
        *,
        cf_data: Optional[TabularData] = None,
        num_history: int = 1,
        is_classification: Optional[bool] = None,
        read_config: Optional[Dict[str, Any]] = None,
        # valid split
        valid_split: Optional[Union[int, float]] = None,
        min_valid_split: int = 100,
        max_valid_split: int = 10000,
        max_valid_split_ratio: float = 0.5,
        valid_split_order: str = "auto",
        # data loader
        shuffle_train: bool = True,
        shuffle_valid: bool = False,
        batch_size: int = 128,
        valid_batch_size: int = 512,
        # inference
        for_inference: bool = False,
    ):
        assert x_train is not None
        self.x_train = x_train
        self.y_train = y_train
        self.x_valid = x_valid
        self.y_valid = y_valid
        self.cf_data = cf_data
        self.num_history = num_history
        if is_classification is None and cf_data is None and not for_inference:
            msg = "`cf_data` should be provided when `is_classification` is None"
            raise ValueError(msg)
        self.is_classification = is_classification
        self.read_config = read_config or {}
        self.valid_split = valid_split
        self.min_valid_split = min_valid_split
        self.max_valid_split = max_valid_split
        self.max_valid_split_ratio = max_valid_split_ratio
        self.valid_split_order = valid_split_order
        self.shuffle_train = shuffle_train
        self.shuffle_valid = shuffle_valid
        self.batch_size = batch_size
        self.valid_batch_size = valid_batch_size

    @property
    def info(self) -> Dict[str, Any]:
        return {
            "cf_data": self.cf_data,
            "input_dim": self.input_dim,
            "num_classes": self.num_classes,
            "num_history": self.num_history,
            "is_classification": self.is_classification,
        }

    def prepare(self, sample_weights: sample_weights_type) -> None:
        self.train_weights, self.valid_weights = _split_sw(sample_weights)
        if self.cf_data is not None:
            self.cf_data.read(self.x_train, self.y_train, **self.read_config)
            if self.is_classification is None:
                self.is_classification = self.cf_data.is_clf
            if self.x_valid is not None:
                self.train_cf_data = self.cf_data
                self.valid_cf_data = self.cf_data.copy_to(self.x_valid, self.y_valid)
            else:
                if isinstance(self.valid_split, int):
                    split = self.valid_split
                else:
                    num_data = len(self.cf_data)
                    if isinstance(self.valid_split, float):
                        split = int(round(self.valid_split * num_data))
                    else:
                        default_split = 0.1
                        num_split = int(round(default_split * num_data))
                        num_split = max(self.min_valid_split, num_split)
                        max_split = int(round(num_data * self.max_valid_split_ratio))
                        max_split = min(max_split, self.max_valid_split)
                        split = min(num_split, max_split)
                if split <= 0:
                    self.train_cf_data = self.cf_data
                    self.valid_cf_data = None
                else:
                    rs = self.cf_data.split(split, order=self.valid_split_order)
                    self.train_cf_data = rs.remained
                    self.valid_cf_data = rs.split
            self.train_data = MLDataset(*self.train_cf_data.processed.xy)
            if self.valid_cf_data is None:
                self.valid_data = None
            else:
                self.valid_data = MLDataset(*self.valid_cf_data.processed.xy)
            self.num_classes = self.train_cf_data.num_classes
            self.input_dim = self.train_cf_data.processed_dim
            return None
        if isinstance(self.x_train, str):
            raise ValueError("`cf_data` should be provided when `x_train` is `str`")
        self.num_classes = None
        self.input_dim = self.x_train.shape[-1]
        self.train_data = MLDataset(self.x_train, self.y_train)
        if self.x_valid is None or self.y_valid is None:
            self.valid_data = None
        else:
            self.valid_data = MLDataset(self.x_valid, self.y_valid)

    def initialize(self) -> Tuple[MLLoader, Optional[MLLoader]]:
        train_loader = MLLoader(
            self.train_data,
            name="train",
            shuffle=self.shuffle_train,
            batch_size=self.batch_size,
            sample_weights=self.train_weights,
        )
        if self.valid_data is None:
            valid_loader = None
        else:
            valid_loader = MLLoader(
                self.valid_data,
                name="valid",
                shuffle=self.shuffle_valid,
                batch_size=self.valid_batch_size,
                sample_weights=self.valid_weights,
            )
        return train_loader, valid_loader

    def _save_info(self, folder: str) -> None:
        info = self.info
        if info["cf_data"] is not None:
            with tempfile.TemporaryDirectory() as tmp_dir:
                tmp_name = os.path.join(tmp_dir, self.tmp_cf_data_name)
                info["cf_data"].save(tmp_name, retain_data=False)
                zip_file = f"{tmp_name}.zip"
                with open(zip_file, "rb") as f:
                    info["cf_data"] = f.read()
                os.remove(zip_file)
        Saving.save_dict(info, self.info_name, folder)

    @classmethod
    def load_info(cls, folder: str) -> Dict[str, Any]:
        d = super().load_info(folder)
        cf_data = d["cf_data"]
        if cf_data is None:
            return d
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_name = os.path.join(tmp_dir, cls.tmp_cf_data_name)
            zip_file = f"{tmp_name}.zip"
            with open(zip_file, "wb") as f:
                f.write(cf_data)
            d["cf_data"] = TabularData.load(tmp_name)
            os.remove(zip_file)
        return d

    @classmethod
    def with_cf_data(
        cls,
        *args: Any,
        is_classification: Optional[bool] = None,
        cf_data_config: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> "MLData":
        if cf_data_config is None:
            cf_data_config = {}
        cf_data_config["default_categorical_process"] = "identical"
        if is_classification is not None:
            cf_data_config["task_type"] = "clf" if is_classification else "reg"
        kwargs["is_classification"] = is_classification
        kwargs["cf_data"] = TabularData(**(cf_data_config or {}))
        return cls(*args, **kwargs)


class MLInferenceData(MLData):
    def __init__(self, x: data_type, y: data_type = None):
        super().__init__(x, y, for_inference=True)


# cv


class CVDataModule(DLDataModule, metaclass=ABCMeta):
    test_transform: Optional[Transforms]


@DLDataModule.register("image_folder")
class ImageFolderData(CVDataModule):
    def __init__(
        self,
        folder: str,
        *,
        batch_size: int,
        num_workers: int = 0,
        shuffle: bool = True,
        drop_train_last: bool = True,
        prefetch_device: Optional[Union[int, str]] = None,
        pin_memory_device: Optional[Union[int, str]] = None,
        extra_label_names: Optional[List[str]] = None,
        transform: Optional[Union[str, List[str], Transforms, Callable]] = None,
        transform_config: Optional[Dict[str, Any]] = None,
        test_shuffle: Optional[bool] = None,
        test_transform: Optional[Union[str, List[str], Transforms, Callable]] = None,
        test_transform_config: Optional[Dict[str, Any]] = None,
        lmdb_config: Optional[Dict[str, Any]] = None,
    ):
        self.folder = folder
        self.shuffle = shuffle
        self.drop_train_last = drop_train_last
        if not torch.cuda.is_available():
            fmt = "cuda is not available but {} is provided, which will have no effect"
            if prefetch_device is not None:
                print(f"{WARNING_PREFIX}{fmt.format('`prefetch_device`')}")
                prefetch_device = None
            if pin_memory_device is not None:
                print(f"{WARNING_PREFIX}{fmt.format('`pin_memory_device`')}")
                pin_memory_device = None
        if prefetch_device is not None:
            if pin_memory_device is None:
                pin_memory_device = prefetch_device
            if pin_memory_device is not None:
                if str(prefetch_device) != str(pin_memory_device):
                    print(
                        f"{WARNING_PREFIX}`prefetch_device` and `pin_memory_device` "
                        "are both provided but they are not consistent, which may "
                        "impact the memory usage and performance"
                    )
        self.prefetch_device = prefetch_device
        self.pin_memory_device = pin_memory_device
        self.extra_label_names = extra_label_names
        self.transform = Transforms.convert(transform, transform_config)
        self.test_shuffle = test_shuffle
        if test_transform is None:
            test_transform = transform
        if test_transform_config is None:
            test_transform_config = transform_config
        self.test_transform = Transforms.convert(test_transform, test_transform_config)
        if self.test_transform is None:
            self.test_transform = self.transform
        self.lmdb_config = lmdb_config
        self.kw = dict(
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=pin_memory_device is not None,
        )

    @property
    def info(self) -> Dict[str, Any]:
        d = shallow_copy_dict(self.kw)
        d["test_shuffle"] = self.test_shuffle
        try:
            json.dumps(self.lmdb_config)
            d["lmdb_config"] = self.lmdb_config
        except Exception as err:
            d["lmdb_config"] = str(err)
        return d

    # TODO : support sample weights
    def prepare(self, sample_weights: sample_weights_type) -> None:
        self.train_data = CVDataset(
            ImageFolderDataset(
                self.folder,
                "train",
                self.transform,
                extra_label_names=self.extra_label_names,
                lmdb_config=self.lmdb_config,
            )
        )
        self.valid_data = CVDataset(
            ImageFolderDataset(
                self.folder,
                "valid",
                self.test_transform,
                extra_label_names=self.extra_label_names,
                lmdb_config=self.lmdb_config,
            )
        )

    def initialize(self) -> Tuple[CVLoader, Optional[CVLoader]]:
        if self.pin_memory_device is not None:
            torch.cuda.set_device(self.pin_memory_device)
        d = shallow_copy_dict(self.kw)
        trd = shallow_copy_dict(d)
        trd["drop_last"] = self.drop_train_last
        kw = {"prefetch_device": self.prefetch_device}
        train_loader = CVLoader(DataLoader(self.train_data, **trd), **kw)  # type: ignore
        d["shuffle"] = self.test_shuffle or self.shuffle
        valid_loader = CVLoader(DataLoader(self.valid_data, **d), **kw)  # type: ignore
        return train_loader, valid_loader

    @staticmethod
    def switch_prefix(src: str, previous: str, now: str) -> None:
        for split in ["train", "valid"]:
            split_folder = os.path.join(src, split)
            for file in os.listdir(split_folder):
                if not file.endswith(".json"):
                    continue
                path = os.path.join(split_folder, file)
                print(f"switching prefix for '{path}'")
                with open(path, "r") as f:
                    rs = json.load(f)
                new_rs: Union[List, Dict[str, Any]]
                if isinstance(rs, list):
                    new_rs = [item.replace(previous, now) for item in rs]
                else:
                    new_rs = {}
                    for k, v in rs.items():
                        new_rs[k.replace(previous, now)] = v.replace(previous, now)
                with open(path, "w", encoding="utf-8") as f:
                    json.dump(new_rs, f, ensure_ascii=False)


class InferenceImageFolderData(CVDataModule):
    def __init__(
        self,
        folder: str,
        *,
        batch_size: int,
        num_workers: int = 0,
        prefetch_device: Optional[Union[int, str]] = None,
        transform: Optional[Union[str, List[str], Transforms, Callable]] = None,
        transform_config: Optional[Dict[str, Any]] = None,
    ):
        self.folder = folder
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.prefetch_device = prefetch_device
        self.transform = Transforms.convert(transform, transform_config)
        self.kw: Dict[str, Any] = dict(batch_size=batch_size, num_workers=num_workers)
        self.prepare(None)

    @property
    def info(self) -> Dict[str, Any]:
        return self.kw

    def prepare(self, sample_weights: sample_weights_type) -> None:
        self.dataset = InferenceImageFolderDataset(self.folder, self.transform)

    def initialize(self) -> Tuple[CVLoader, Optional[CVLoader]]:
        args = self.batch_size, self.num_workers, self.prefetch_device
        loader = self.dataset.make_loader(*args)
        return loader, None


class PrepareResults(NamedTuple):
    data: ImageFolderData
    tgt_folder: str


class _PreparationProtocol:
    @property
    def extra_labels(self) -> Optional[List[str]]:
        pass

    def prepare_src_folder(self, src_path: str) -> None:
        pass

    def filter(self, hierarchy: List[str]) -> bool:
        pass

    def get_label(self, hierarchy: List[str]) -> Any:
        pass

    def get_extra_label(self, label_name: str, hierarchy: List[str]) -> Any:
        pass

    def copy(self, src_path: str, tgt_path: str) -> None:
        pass

    def get_new_img_path(self, idx: int, split_folder: str, old_img_path: str) -> str:
        pass

    def is_ready(self, tgt_folder: str) -> bool:
        for split in ["train", "valid"]:
            extra_keys = [f"{key}_labels" for key in self.extra_labels or []]
            for key in [LABEL_KEY] + extra_keys:
                path = os.path.join(tgt_folder, split, f"{key}.json")
                if not os.path.isfile(path):
                    return False
        return True

    def get_num_classes(self, tgt_folder: str) -> Dict[str, int]:
        num_classes = {}
        for label_name in [LABEL_KEY] + (self.extra_labels or []):
            path = os.path.join(tgt_folder, f"idx2{label_name}.json")
            if not os.path.isfile(path):
                num_classes[label_name] = 0
                continue
            with open(path, "r") as f:
                num_classes[label_name] = len(json.load(f))
        return num_classes


class DefaultPreparation(_PreparationProtocol):
    @property
    def extra_labels(self) -> Optional[List[str]]:
        return None

    def filter(self, hierarchy: List[str]) -> bool:
        return True

    def get_label(self, hierarchy: List[str]) -> Any:
        return 0

    def copy(self, src_path: str, tgt_path: str) -> None:
        shutil.copyfile(src_path, tgt_path)

    def get_new_img_path(self, idx: int, split_folder: str, old_img_path: str) -> str:
        ext = os.path.splitext(old_img_path)[1]
        return os.path.join(split_folder, f"{idx}{ext}")


def prepare_image_folder(
    src_folder: str,
    tgt_folder: str,
    *,
    to_index: bool,
    prefix: Optional[str] = None,
    preparation: _PreparationProtocol = DefaultPreparation(),
    force_rerun: bool = False,
    extensions: Optional[Set[str]] = None,
    make_labels_in_parallel: bool = False,
    num_jobs: int = 8,
    train_all_data: bool = False,
    valid_split: Union[int, float] = 0.1,
    max_num_valid: int = 10000,
    lmdb_config: Optional[Dict[str, Any]] = None,
    use_tqdm: bool = True,
) -> str:
    if prefix is not None:
        src_folder = os.path.join(prefix, src_folder)
        tgt_folder = os.path.join(prefix, tgt_folder)

    if not force_rerun and preparation.is_ready(tgt_folder):
        return tgt_folder

    preparation.prepare_src_folder(src_folder)
    if os.path.isdir(tgt_folder):
        print(f"{WARNING_PREFIX}'{tgt_folder}' already exists, it will be removed")
        shutil.rmtree(tgt_folder)
    os.makedirs(tgt_folder, exist_ok=True)

    print(f"{INFO_PREFIX}collecting hierarchies")

    def hierarchy_callback(hierarchy: List[str], path: str) -> None:
        hierarchy = hierarchy[prefix_idx:]
        if not preparation.filter(hierarchy):
            return None
        hierarchy_list.append(hierarchy)
        all_img_paths.append(path)

    all_img_paths: List[str] = []
    hierarchy_list: List[List[str]] = []
    if extensions is None:
        extensions = {".jpg", ".png"}
    prefix_idx = 0
    if prefix is not None:
        prefix_idx = len(prefix.split(os.path.sep))
    walk(src_folder, hierarchy_callback, extensions)

    def get_labels(
        label_fn: Callable,
        label_name: Optional[str] = None,
    ) -> List[Any]:
        def task(h: List[str]) -> Any:
            try:
                args = (h,) if label_name is None else (label_name, h)
                return label_fn(*args)
            except Exception as err:
                err_path = "/".join(h)
                msg = f"error occurred ({err}) when getting label of {err_path}"
                print(f"{ERROR_PREFIX}{msg}")
                return None

        if not make_labels_in_parallel:
            return [task(h) for h in tqdm(hierarchy_list)]
        parallel = Parallel(num_jobs, use_tqdm=use_tqdm)
        num_files = len(hierarchy_list)
        random_indices = np.random.permutation(num_files).tolist()
        shuffled = [hierarchy_list[i] for i in random_indices]
        groups = parallel.grouped(task, shuffled).ordered_results
        shuffled_results: List[Any] = sum(groups, [])
        final_results = [None] * num_files
        for idx, rs in zip(random_indices, shuffled_results):
            final_results[idx] = rs
        return final_results

    print(f"{INFO_PREFIX}making labels")
    labels = get_labels(preparation.get_label)
    excluded_indices = {i for i, label in enumerate(labels) if label is None}
    extra_labels_dict: Optional[Dict[str, List[str]]] = None
    extra_labels = preparation.extra_labels
    extra_label_fn = preparation.get_extra_label
    if extra_labels is not None:
        extra_labels_dict = {}
        print(f"{INFO_PREFIX}making extra labels")
        for el_name in extra_labels:
            extra_labels_dict[el_name] = get_labels(extra_label_fn, el_name)
        for extra_labels in extra_labels_dict.values():
            for i, extra_label in enumerate(extra_labels):
                if extra_label is None:
                    excluded_indices.add(i)

    def get_raw_2idx(raw_labels: List[Any]) -> Dict[Any, Any]:
        return {
            v: numpy_token if isinstance(v, str) and v.endswith(".npy") else v
            for v in raw_labels
        }

    def check_dump_mappings(l2i: Dict[Any, Any]) -> bool:
        all_indices = set(l2i.values())
        if len(all_indices) > 1:
            return True
        return list(all_indices)[0] != numpy_token

    numpy_token = "[NUMPY]"
    if to_index:
        label2idx = {label: i for i, label in enumerate(sorted(set(labels)))}
        labels_dict = {"": [label2idx[label] for label in labels]}
        dump_mappings = True
    else:
        labels_dict = {"": labels}
        label2idx = get_raw_2idx(sorted(set(labels)))
        dump_mappings = check_dump_mappings(label2idx)

    open_file_from = lambda folder: lambda file: open(
        os.path.join(folder, file), "w", encoding="utf-8"
    )
    open_tgt_file = open_file_from(tgt_folder)

    if dump_mappings:
        with open_tgt_file(f"{LABEL_KEY}2idx.json") as f:
            json.dump(label2idx, f, ensure_ascii=False)
        with open_tgt_file(f"idx2{LABEL_KEY}.json") as f:
            json.dump({v: k for k, v in label2idx.items()}, f, ensure_ascii=False)

    if extra_labels_dict is not None:
        for el_name, label_collection in extra_labels_dict.items():
            if not to_index:
                labels_dict[el_name] = label_collection  # type: ignore
                extra2idx = get_raw_2idx(sorted(set(label_collection)))
                dump_mappings = check_dump_mappings(extra2idx)
            else:
                extra2idx = {
                    extra_label: i  # type: ignore
                    for i, extra_label in enumerate(sorted(set(label_collection)))
                }
                labels_dict[el_name] = [extra2idx[el] for el in label_collection]
                dump_mappings = True
            if dump_mappings:
                with open_tgt_file(f"{el_name}2idx.json") as f:
                    json.dump(extra2idx, f, ensure_ascii=False)
                with open_tgt_file(f"idx2{el_name}.json") as f:
                    eld = {v: k for k, v in extra2idx.items()}
                    json.dump(eld, f, ensure_ascii=False)

    # exclude samples
    if excluded_indices:
        print(f"{WARNING_PREFIX}{len(excluded_indices)} samples will be excluded")
    for i in sorted(excluded_indices)[::-1]:
        for sub_labels in labels_dict.values():
            sub_labels.pop(i)
        all_img_paths.pop(i)

    # prepare core
    num_sample = len(all_img_paths)
    if valid_split < 1:
        valid_split = min(max_num_valid, int(round(num_sample * valid_split)))
    assert isinstance(valid_split, int)

    train_portion = (num_sample - valid_split) / num_sample
    label_indices_mapping: Dict[Any, List[int]] = {}
    for i, label in enumerate(labels):
        if isinstance(label, str) and label.endswith(".npy"):
            label = numpy_token
        label_indices_mapping.setdefault(label, []).append(i)
    tuple(map(random.shuffle, label_indices_mapping.values()))
    train_indices_list: List[List[int]] = []
    valid_indices_list: List[List[int]] = []
    for label_indices in label_indices_mapping.values():
        num_label_samples = len(label_indices)
        num_train = int(round(train_portion * num_label_samples))
        num_train = min(num_train, num_label_samples - 1)
        if num_train == 0:
            train_indices_list.append([label_indices[0]])
        else:
            train_indices_list.append(label_indices[:num_train])
        valid_indices_list.append(label_indices[num_train:])

    def propagate(src: List[List[int]], tgt: List[List[int]]) -> None:
        resolved = 0
        src_lengths = list(map(len, src))
        sorted_indices = np.argsort(src_lengths).tolist()[::-1]
        while True:
            for idx in sorted_indices:
                if len(src[idx]) > 1:
                    tgt[idx].append(src[idx].pop())
                resolved += 1
                if resolved == diff:
                    break
            if resolved == diff:
                break

    diff = sum(map(len, valid_indices_list)) - valid_split
    if diff > 0:
        propagate(valid_indices_list, train_indices_list)
    elif diff < 0:
        diff *= -1
        propagate(train_indices_list, valid_indices_list)
    merged_train_indices: List[int] = sorted(set(sum(train_indices_list, [])))
    merged_valid_indices: List[int] = sorted(set(sum(valid_indices_list, [])))
    if train_all_data:
        merged_train_indices.extend(merged_valid_indices)
    train_indices = np.array(merged_train_indices)
    valid_indices = np.array(merged_valid_indices)

    def save(indices: np.ndarray, d_num_jobs: int, dtype: str) -> None:
        def record(idx: int) -> Optional[Tuple[str, Dict[str, Any]]]:
            split_folder = os.path.join(tgt_folder, dtype)
            os.makedirs(split_folder, exist_ok=True)
            img_path = all_img_paths[idx]
            new_img_path = preparation.get_new_img_path(idx, split_folder, img_path)
            try:
                preparation.copy(img_path, new_img_path)
                key = os.path.abspath(new_img_path)
                idx_labels: Dict[str, Any] = {}
                for label_t, t_labels in labels_dict.items():
                    idx_labels[label_t] = {key: t_labels[idx]}
                return key, idx_labels
            except Exception as err:
                print(f"error occurred with {img_path} : {err}")
                return None

        parallel = Parallel(d_num_jobs, use_tqdm=use_tqdm)
        results: List[Tuple[str, Dict[str, Any]]]
        indices = indices.copy()
        np.random.shuffle(indices)
        results = sum(parallel.grouped(record, indices).ordered_results, [])
        d_valid_indices = [i for i, r in enumerate(results) if r is not None]
        results = [results[i] for i in d_valid_indices]
        valid_paths = [all_img_paths[idx] for idx in indices[d_valid_indices]]
        new_paths, all_labels_list = zip(*results)
        merged_labels = shallow_copy_dict(all_labels_list[0])
        for sub_labels_ in all_labels_list[1:]:
            for k, v in shallow_copy_dict(sub_labels_).items():
                merged_labels[k].update(v)
        print(
            "\n".join(
                [
                    "",
                    "=" * 100,
                    f"num {dtype} samples : {len(next(iter(merged_labels.values())))}",
                    f"num {dtype} label types : {len(merged_labels)}",
                    "-" * 100,
                    "",
                ]
            )
        )
        open_dtype_file = open_file_from(os.path.join(tgt_folder, dtype))

        with open_dtype_file("paths.json") as f_:
            json.dump(new_paths, f_, ensure_ascii=False)
        path_mapping = dict(zip(new_paths, valid_paths))
        with open_dtype_file("path_mapping.json") as f_:
            json.dump(path_mapping, f_, ensure_ascii=False)
        for label_type, type_labels in merged_labels.items():
            delim = "_" if label_type else ""
            label_file = f"{label_type}{delim}{LABEL_KEY}.json"
            with open_dtype_file(label_file) as f_:
                json.dump(type_labels, f_, ensure_ascii=False)
        # lmdb
        if lmdb_config is None:
            return None
        local_lmdb_config = shallow_copy_dict(lmdb_config)
        local_lmdb_config.setdefault("path", default_lmdb_path(tgt_folder, dtype))
        local_lmdb_config.setdefault("map_size", 1099511627776 * 2)
        db = lmdb.open(**local_lmdb_config)
        context = db.begin(write=True)
        d_num_samples = len(results)
        iterator = zip(range(d_num_samples), new_paths, all_labels_list)
        if use_tqdm:
            iterator = tqdm(iterator, total=d_num_samples, desc="lmdb")
        for i, path, i_labels in iterator:
            i_new_labels = {}
            for k, v in i_labels.items():
                vv = v[path]
                if isinstance(vv, str):
                    if vv.endswith(".npy"):
                        vv = np.load(vv)
                i_new_labels[k] = vv
            context.put(
                str(i).encode("ascii"),
                dill.dumps(LMDBItem(np.array(Image.open(path)), i_new_labels)),
            )
        context.put(
            "length".encode("ascii"),
            str(d_num_samples).encode("ascii"),
        )
        context.commit()
        db.sync()
        db.close()

    save(train_indices, max(1, num_jobs), "train")
    save(valid_indices, max(1, num_jobs // 2), "valid")
    return tgt_folder


def prepare_image_folder_data(
    src_folder: str,
    tgt_folder: str,
    *,
    to_index: bool,
    batch_size: int,
    prefix: Optional[str] = None,
    preparation: _PreparationProtocol = DefaultPreparation(),
    num_workers: int = 0,
    shuffle: bool = True,
    drop_train_last: bool = True,
    prefetch_device: Optional[Union[int, str]] = None,
    pin_memory_device: Optional[Union[int, str]] = None,
    transform: Optional[Union[str, List[str], Transforms, Callable]] = None,
    transform_config: Optional[Dict[str, Any]] = None,
    test_shuffle: Optional[bool] = None,
    test_transform: Optional[Union[str, List[str], Transforms, Callable]] = None,
    test_transform_config: Optional[Dict[str, Any]] = None,
    train_all_data: bool = False,
    force_rerun: bool = False,
    extensions: Optional[Set[str]] = None,
    make_labels_in_parallel: bool = False,
    num_jobs: int = 8,
    valid_split: Union[int, float] = 0.1,
    max_num_valid: int = 10000,
    lmdb_config: Optional[Dict[str, Any]] = None,
    use_tqdm: bool = True,
) -> PrepareResults:
    tgt_folder = prepare_image_folder(
        src_folder,
        tgt_folder,
        to_index=to_index,
        prefix=prefix,
        preparation=preparation,
        force_rerun=force_rerun,
        extensions=extensions,
        make_labels_in_parallel=make_labels_in_parallel,
        num_jobs=num_jobs,
        train_all_data=train_all_data,
        valid_split=valid_split,
        max_num_valid=max_num_valid,
        lmdb_config=lmdb_config,
        use_tqdm=use_tqdm,
    )
    data = ImageFolderData(
        tgt_folder,
        batch_size=batch_size,
        num_workers=num_workers,
        shuffle=shuffle,
        drop_train_last=drop_train_last,
        prefetch_device=prefetch_device,
        pin_memory_device=pin_memory_device,
        extra_label_names=preparation.extra_labels,
        transform=transform,
        transform_config=transform_config,
        test_shuffle=test_shuffle,
        test_transform=test_transform,
        test_transform_config=test_transform_config,
        lmdb_config=lmdb_config,
    )
    return PrepareResults(data, tgt_folder)


__all__ = [
    "TensorData",
    "TensorDictData",
    "DummyData",
    "MLData",
    "MLInferenceData",
    "CVDataModule",
    "ImageFolderData",
    "InferenceImageFolderData",
    "DefaultPreparation",
    "prepare_image_folder",
    "prepare_image_folder_data",
]



import torch

import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import List
from typing import Callable
from typing import Optional

from .....modules.blocks import Conv2d


def conv3x3(
    in_channels: int,
    out_channels: int,
    stride: int = 1,
    groups: int = 1,
    dilation: int = 1,
) -> Conv2d:
    return Conv2d(
        in_channels,
        out_channels,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv1x1(in_channels: int, out_channels: int, stride: int = 1) -> Conv2d:
    return Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[[int], nn.Module]] = None,
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")

        self.conv1 = conv3x3(in_channels, out_channels, stride)
        self.bn1 = norm_layer(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(out_channels, out_channels)
        self.bn2 = norm_layer(out_channels)
        self.downsample = downsample
        self.stride = stride

    def forward(self, net: Tensor) -> Tensor:
        inp = net

        net = self.conv1(net)
        net = self.bn1(net)
        net = self.relu(net)

        net = self.conv2(net)
        net = self.bn2(net)

        if self.downsample is not None:
            inp = self.downsample(inp)
        net = net + inp
        net = self.relu(net)

        return net


class ResNet(nn.Module):
    def __init__(
        self,
        layers: List[int],
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: Optional[List[bool]] = None,
        norm_layer: Optional[Callable[[int], nn.Module]] = None,
    ):
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer
        self.num_channels = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be None "
                "or a 3-element tuple, got {}".format(replace_stride_with_dilation)
            )
        self.groups = groups
        self.base_width = width_per_group
        self.conv1 = Conv2d(
            3,
            self.num_channels,
            kernel_size=7,
            stride=1,
            padding=3,
            bias=False,
        )
        self.bn1 = norm_layer(self.num_channels)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool1 = nn.MaxPool2d(
            kernel_size=3,
            stride=2,
            padding=1,
            return_indices=True,
        )
        self.maxpool2 = nn.MaxPool2d(
            kernel_size=3,
            stride=2,
            padding=1,
            return_indices=True,
        )
        self.maxpool3 = nn.MaxPool2d(
            kernel_size=3,
            stride=2,
            padding=1,
            return_indices=True,
        )
        self.maxpool4 = nn.MaxPool2d(
            kernel_size=3,
            stride=2,
            padding=1,
            return_indices=True,
        )
        self.maxpool5 = nn.MaxPool2d(
            kernel_size=3,
            stride=2,
            padding=1,
            return_indices=True,
        )
        self.layer1 = self._make_layer(64, layers[0])
        self.layer2 = self._make_layer(
            128,
            layers[1],
            stride=1,
            dilate=replace_stride_with_dilation[0],
        )
        self.layer3 = self._make_layer(
            256,
            layers[2],
            stride=1,
            dilate=replace_stride_with_dilation[1],
        )
        self.layer4 = self._make_layer(
            512,
            layers[3],
            stride=1,
            dilate=replace_stride_with_dilation[2],
        )

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512, 1000)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(
        self,
        num_channels: int,
        num_blocks: int,
        stride: int = 1,
        dilate: bool = False,
    ) -> nn.Sequential:
        downsample = None
        norm_layer = self._norm_layer
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.num_channels != num_channels:
            downsample = nn.Sequential(
                conv1x1(self.num_channels, num_channels, stride),
                norm_layer(num_channels),
            )

        layers = [
            BasicBlock(
                self.num_channels,
                num_channels,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
            )
        ]
        self.num_channels = num_channels
        for _ in range(1, num_blocks):
            layers.append(
                BasicBlock(
                    self.num_channels,
                    num_channels,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def forward(self, net: Tensor) -> Tensor:
        x1 = self.conv1(net)
        x1 = self.bn1(x1)
        x1 = self.relu(x1)
        x1, idx1 = self.maxpool1(x1)

        x2, idx2 = self.maxpool2(x1)
        x2 = self.layer1(x2)

        x3, idx3 = self.maxpool3(x2)
        x3 = self.layer2(x3)

        x4, idx4 = self.maxpool4(x3)
        x4 = self.layer3(x4)

        x5, idx5 = self.maxpool5(x4)
        x5 = self.layer4(x5)

        x_cls = self.avgpool(x5)
        x_cls = torch.flatten(x_cls, 1)
        x_cls = self.fc(x_cls)

        return x_cls


def resnet34_mp(**kwargs: Any) -> ResNet:
    return ResNet([3, 4, 6, 3], **kwargs)


__all__ = ["resnet34_mp"]



import dill

import numpy as np

from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Union
from cftool.misc import WithRegister

from ....misc.toolkit import check_available
from ....misc.toolkit import download_tokenizer
from ....misc.toolkit import get_compatible_name


tokenizers: Dict[str, Type["TokenizerProtocol"]] = {}


class TokenizerProtocol(WithRegister["TokenizerProtocol"], metaclass=ABCMeta):
    d = tokenizers

    @abstractmethod
    def tokenize(self, texts: Union[str, List[str]]) -> np.ndarray:
        pass

    @classmethod
    def make(cls, name: str, config: Dict[str, Any]) -> "TokenizerProtocol":
        tag = "tokenizers"
        repo = "pretrained-models"
        name = get_compatible_name(tag, repo, name, [(3, 8), (3, 9)])
        if check_available(tag, repo, name):
            with open(download_tokenizer(name), "rb") as f:
                return dill.load(f)
        return super().make(name, config)


__all__ = [
    "TokenizerProtocol",
]



from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from functools import partial
from torchvision.datasets import MNIST

from ....data import Transforms
from ....types import tensor_dict_type
from ....types import sample_weights_type
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....constants import DATA_CACHE_DIR
from ....constants import ORIGINAL_LABEL_KEY
from ....data import CVDataset
from ....data import CVLoader
from ....data import DataLoader
from ....data import DLDataModule
from ....data.interface import CVDataModule


def batch_callback(
    label_callback: Optional[Callable[[Tuple[Tensor, Tensor]], Tensor]],
    batch: Tuple[Tensor, Tensor],
) -> tensor_dict_type:
    img, labels = batch
    if label_callback is None:
        actual_labels = labels.view(-1, 1)
    else:
        actual_labels = label_callback(batch)
    return {
        INPUT_KEY: img,
        LABEL_KEY: actual_labels,
        ORIGINAL_LABEL_KEY: labels,
    }


@DLDataModule.register("mnist")
class MNISTData(CVDataModule):
    def __init__(
        self,
        *,
        root: str = DATA_CACHE_DIR,
        shuffle: bool = True,
        batch_size: int = 64,
        num_workers: int = 0,
        drop_train_last: bool = True,
        transform: Optional[Union[str, List[str], Transforms, Callable]],
        transform_config: Optional[Dict[str, Any]] = None,
        label_callback: Optional[Callable[[Tuple[Tensor, Tensor]], Tensor]] = None,
    ):
        self.root = root
        self.shuffle = shuffle
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.drop_train_last = drop_train_last
        self.transform = Transforms.convert(transform, transform_config)
        self.test_transform = self.transform
        self.label_callback = label_callback

    @property
    def info(self) -> Dict[str, Any]:
        return dict(root=self.root, shuffle=self.shuffle, batch_size=self.batch_size)

    # TODO : support sample weights
    def prepare(self, sample_weights: sample_weights_type) -> None:
        self.train_data = CVDataset(
            MNIST(
                self.root,
                transform=self.transform,
                download=True,
            )
        )
        self.valid_data = CVDataset(
            MNIST(
                self.root,
                train=False,
                transform=self.transform,
                download=True,
            )
        )

    def initialize(self) -> Tuple[CVLoader, Optional[CVLoader]]:
        train_loader = CVLoader(
            DataLoader(
                self.train_data,
                batch_size=self.batch_size,
                shuffle=self.shuffle,
                num_workers=self.num_workers,
                drop_last=self.drop_train_last,
            ),
            partial(batch_callback, self.label_callback),
        )
        valid_loader = CVLoader(
            DataLoader(
                self.valid_data,
                batch_size=self.batch_size,
                shuffle=self.shuffle,
                num_workers=self.num_workers,
            ),
            partial(batch_callback, self.label_callback),
        )
        return train_loader, valid_loader


__all__ = [
    "MNISTData",
]
